from kgb import SpyAgency

from reviewboard.diffviewer.chunk_generator import DiffChunkGenerator
from reviewboard.diffviewer.settings import DiffSettings
from reviewboard.scmtools.core import PRE_CREATION
from reviewboard.testing import TestCase


class DiffChunkGeneratorTests(SpyAgency, TestCase):
    """Unit tests for DiffChunkGenerator."""

    fixtures = ['test_scmtools']

    COMMIT_1_DIFF = (
        b'diff --git a/README b/README\n'
        b'index 94bdd3e..197009f 100644\n'
        b'--- README\n'
        b'+++ README\n'
        b'@@ -1,1 +1,1 @@\n'
        b'-Hello, world!\n'
        b'+Hi, world!\n'
    )

    COMMIT_2_DIFF = (
        b'diff --git a/README b/README\n'
        b'index 197009f..87abad9 100644\n'
        b'--- README\n'
        b'+++ README\n'
        b'@@ -1,1 +1,1 @@\n'
        b'-Hi, world!\n'
        b'+Yo, world.\n'
    )

    COMMIT_3_DIFF = (
        b'diff --git a/README b/README\n'
        b'index 87abad9..fe1678a 100644\n'
        b'--- README\n'
        b'+++ README\n'
        b'@@ -1,1 +1,1 @@\n'
        b'-Yo, world.\n'
        b'+Yo, dog.\n'
    )

    COMMIT_1_2_SQUASHED_DIFF = (
        b'diff --git a/README b/README\n'
        b'index 94bdd3e..87abad9 100644\n'
        b'--- README\n'
        b'+++ README\n'
        b'@@ -1,1 +1,1 @@\n'
        b'-Hello, world!\n'
        b'+Yo, world.\n'
    )

    COMMIT_1_3_SQUASHED_DIFF = (
        b'diff --git a/README b/README\n'
        b'index 94bdd3e..fe1678a 100644\n'
        b'--- README\n'
        b'+++ README\n'
        b'@@ -1,1 +1,1 @@\n'
        b'-Hello, world!\n'
        b'+Yo, dog.\n'
    )

    def setUp(self):
        super(DiffChunkGeneratorTests, self).setUp()

        self.repository = self.create_repository(tool_name='Test')
        self.diffset = self.create_diffset(repository=self.repository)
        self.filediff = self.create_filediff(diffset=self.diffset)
        self.generator = DiffChunkGenerator(
            None, self.filediff, diff_settings=DiffSettings.create())

    def test_get_chunks_with_empty_added_file(self):
        """Testing DiffChunkGenerator.get_chunks with empty added file"""
        self.filediff.source_revision = PRE_CREATION
        self.filediff.extra_data.update({
            'raw_insert_count': 0,
            'raw_delete_count': 0,
        })

        self.assertEqual(len(list(self.generator.get_chunks())), 0)

    def test_get_chunks_with_explicit_encoding(self):
        """Testing DiffChunkGenerator.get_chunks with explicit encoding on
        FileDiff
        """
        self.filediff.diff = (
            b'--- README\n'
            b'+++ README\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-%s\n'
            b'+%s\n'
            % ('Hello, world!'.encode('utf-16'),
               'Hi, everybody!'.encode('utf-16'))
        )
        self.filediff.source_file = '/test-file;encoding=utf-16'
        self.filediff.extra_data['encoding'] = 'utf-16'

        self.spy_on(self.repository.get_file)

        chunks = list(self.generator.get_chunks())
        self.assertEqual(len(chunks), 1)
        self.assertEqual(
            chunks[0],
            {
                'change': 'replace',
                'collapsable': False,
                'index': 0,
                'lines': [
                    [
                        1, 1,
                        'Hello, world!',
                        None,
                        1,
                        'Hi, everybody!',
                        None,
                        False,
                    ]
                ],
                'meta': {
                    'left_headers': [],
                    'right_headers': [],
                    'whitespace_chunk': False,
                    'whitespace_lines': [],
                },
                'numlines': 1,
            })

        self.assertTrue(self.repository.get_file.last_returned(
            'Hello, world!\n'.encode('utf-16')))

    def test_get_chunks_with_replace_in_added_file_with_parent_diff(self):
        """Testing DiffChunkGenerator.get_chunks with replace chunks in
        added file with parent diff
        """
        self.filediff.diff = (
            b'--- README\n'
            b'+++ README\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-line\n'
            b'+line.\n'
        )
        self.filediff.parent_diff = (
            b'--- README\n'
            b'+++ README\n'
            b'@@ -0,0 +1,1 @@\n'
            b'+line\n'
        )
        self.filediff.source_revision = PRE_CREATION
        self.filediff.extra_data.update({
            'raw_insert_count': 1,
            'raw_delete_count': 1,
            'insert_count': 0,
            'delete_count': 0,
        })

        self.assertEqual(len(list(self.generator.get_chunks())), 1)

    def test_get_chunks_with_commit_and_base_filediff(self):
        """Testing DiffChunkGenerator.get_chunks with commit using base_filediff
        """
        commit1 = self.create_diffcommit(
            commit_id='abc1234',
            parent_id='abc1233',
            diffset=self.diffset,
            diff_contents=self.COMMIT_1_DIFF)
        commit2 = self.create_diffcommit(
            commit_id='abc1235',
            parent_id='abc1234',
            diffset=self.diffset,
            diff_contents=self.COMMIT_2_DIFF)

        self.diffset.finalize_commit_series(
            cumulative_diff=self.COMMIT_1_2_SQUASHED_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        base_filediff = commit1.files.get()
        tip_filediff = commit2.files.get()

        generator = DiffChunkGenerator(request=None,
                                       filediff=tip_filediff,
                                       base_filediff=base_filediff,
                                       diff_settings=DiffSettings.create())

        chunks = list(generator.get_chunks())
        self.assertEqual(len(chunks), 1)

        chunk = chunks[0]
        self.assertEqual(chunk['index'], 0)
        self.assertEqual(chunk['change'], 'replace')
        self.assertEqual(
            chunk['lines'],
            [[
                1,
                1, 'Hi, world!', [(0, 2), (9, 10)],
                1, 'Yo, world.', [(0, 2), (9, 10)],
                False,
            ]])

    def test_get_chunks_with_commit_and_no_base_filediff(self):
        """Testing DiffChunkGenerator.get_chunks with commit and no
        base_filediff
        """
        self.create_diffcommit(
            commit_id='abc1234',
            parent_id='abc1233',
            diffset=self.diffset,
            diff_contents=self.COMMIT_1_DIFF)
        self.create_diffcommit(
            commit_id='abc1235',
            parent_id='abc1234',
            diffset=self.diffset,
            diff_contents=self.COMMIT_2_DIFF)
        commit3 = self.create_diffcommit(
            commit_id='abc1236',
            parent_id='abc1235',
            diffset=self.diffset,
            diff_contents=self.COMMIT_3_DIFF)

        self.diffset.finalize_commit_series(
            cumulative_diff=self.COMMIT_1_3_SQUASHED_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        filediff = commit3.files.get()

        generator = DiffChunkGenerator(request=None,
                                       filediff=filediff,
                                       diff_settings=DiffSettings.create())

        chunks = list(generator.get_chunks())
        self.assertEqual(len(chunks), 1)

        chunk = chunks[0]
        self.assertEqual(chunk['index'], 0)
        self.assertEqual(chunk['change'], 'replace')
        self.assertEqual(
            chunk['lines'],
            [[
                1,
                1, 'Hello, world!', None,
                1, 'Yo, dog.', None,
                False,
            ]])

    def test_get_chunks_with_commit_and_base_tip_same(self):
        """Testing DiffChunkGenerator.get_chunks with commit and base_filediff
        same as filediff
        """
        commit = self.create_diffcommit(
            commit_id='abc1234',
            parent_id='abc1233',
            diffset=self.diffset,
            diff_contents=self.COMMIT_1_DIFF)

        self.diffset.finalize_commit_series(
            cumulative_diff=self.COMMIT_1_DIFF,
            validation_info=None,
            validate=False,
            save=True)

        filediff = commit.files.get()

        generator = DiffChunkGenerator(request=None,
                                       filediff=filediff,
                                       base_filediff=filediff,
                                       diff_settings=DiffSettings.create())

        chunks = list(generator.get_chunks())
        self.assertEqual(len(chunks), 1)

        chunk = chunks[0]
        self.assertEqual(chunk['index'], 0)
        self.assertEqual(chunk['change'], 'equal')
        self.assertEqual(
            chunk['lines'],
            [[
                1,
                1, 'Hi, world!', [],
                1, 'Hi, world!', [],
                False,
            ]])

    def test_get_chunks_with_commit_and_file_recreated(self):
        """Testing DiffChunkGenerator.get_chunks with commit and file recreated
        in prior commit
        """
        commit1, commit2 = self._make_delete_recreate_commits()
        filediff = commit2.files.get()

        generator = DiffChunkGenerator(request=None,
                                       filediff=filediff,
                                       diff_settings=DiffSettings.create())

        chunks = list(generator.get_chunks())
        self.assertEqual(len(chunks), 1)

        chunk = chunks[0]
        self.assertEqual(chunk['index'], 0)
        self.assertEqual(chunk['change'], 'insert')
        self.assertEqual(
            chunk['lines'],
            [[
                1,
                '', '', [],
                1, 'This is a new file.', [],
                False,
            ]])

    def test_get_chunks_with_commit_and_file_recreated_and_base_deleted(self):
        """Testing DiffChunkGenerator.get_chunks with commit and file recreated
        in prior commit with base_filediff as deleted file
        """
        commit1, commit2 = self._make_delete_recreate_commits()
        base_filediff = commit1.files.get()
        filediff = commit2.files.get()

        generator = DiffChunkGenerator(request=None,
                                       filediff=filediff,
                                       base_filediff=base_filediff,
                                       diff_settings=DiffSettings.create())

        chunks = list(generator.get_chunks())
        self.assertEqual(len(chunks), 1)

        chunk = chunks[0]
        self.assertEqual(chunk['index'], 0)
        self.assertEqual(chunk['change'], 'insert')
        self.assertEqual(
            chunk['lines'],
            [[
                1,
                '', '', [],
                1, 'This is a new file.', [],
                False,
            ]])

    def test_line_counts_unmodified_by_interdiff(self):
        """Testing that line counts are not modified by interdiffs where the
        changes are reverted
        """
        self.filediff.source_revision = PRE_CREATION
        self.filediff.diff = (
            b'--- README\n'
            b'+++ README\n'
            b'@@ -0,0 +1,1 @@\n'
            b'+line\n'
        )

        # We have to consume everything from the get_chunks generator in order
        # for the line counts to be set on the FileDiff.
        self.assertEqual(len(list(self.generator.get_chunks())), 1)

        line_counts = self.filediff.get_line_counts()

        # Simulate an interdiff where the changes are reverted.
        interdiff_generator = DiffChunkGenerator(
            request=None,
            filediff=self.filediff,
            interfilediff=None,
            force_interdiff=True,
            diff_settings=DiffSettings.create())

        # Again, just consuming the generator.
        self.assertEqual(len(list(interdiff_generator.get_chunks())), 1)

        self.assertEqual(line_counts, self.filediff.get_line_counts())

    def _make_delete_recreate_commits(self):
        """Finalize and return commits for a delete/re-create test.

        This creates two commits. In the first one, an upstream file is
        deleted. In the second, it's re-created.

        The commits are finalized before being returned.

        Returns:
            list of reviewboard.diffviewer.models.diffcommit.DiffCommit:
            The newly-created commits.
        """
        commits = [
            self.create_diffcommit(
                commit_id='abc1234',
                parent_id='abc1233',
                diffset=self.diffset,
                diff_contents=(
                    b'diff --git a/README b/README\n'
                    b'deleted file mode 100644\n'
                    b'index 94bdd3e..0000000\n'
                    b'--- README\n'
                    b'+++ /dev/null\n'
                    b'@@ -1,1 +0,0 @@\n'
                    b'-Hello, world!\n'
                )),
            self.create_diffcommit(
                commit_id='abc1235',
                parent_id='abc1234',
                diffset=self.diffset,
                diff_contents=(
                    b'diff --git a/README b/README\n'
                    b'new file mode 100644\n'
                    b'index 0000000..ba178ca\n'
                    b'--- /dev/null\n'
                    b'+++ README\n'
                    b'@@ -0,0 +1,1 @@\n'
                    b'+This is a new file.\n'
                )),
        ]

        self.diffset.finalize_commit_series(
            cumulative_diff=(
                b'diff --git a/README b/README\n'
                b'index 94bdd3e..0000000 10644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hello, world!\n'
                b'+This is a new file.\n'
            ),
            validation_info=None,
            validate=False,
            save=True)

        return commits



"""Unit tests for reviewboard.accounts.middleware.x509_auth_middleware."""

from django.contrib.auth.models import AnonymousUser
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import HttpResponse
from django.test.client import RequestFactory
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.accounts.backends import X509Backend
from reviewboard.accounts.middleware import x509_auth_middleware
from reviewboard.testing import TestCase


class X509AuthMiddlewareTests(TestCase):
    """Unit tests for reviewboard.accounts.middleware.x509_auth_middleware."""

    fixtures = ['test_users']

    def setUp(self):
        super(X509AuthMiddlewareTests, self).setUp()

        self.middleware = SessionMiddleware(
            x509_auth_middleware(
                lambda request: HttpResponse('')))
        self.siteconfig = SiteConfiguration.objects.get_current()

        self.enabled_settings = {
            'auth_backend': X509Backend.backend_id,
        }

        self.request = RequestFactory().get('/')
        self.request.user = AnonymousUser()
        self.request.is_secure = lambda: True

    def test_process_request_without_enabled(self):
        """Testing x509_auth_middleware without backend enabled
        """
        self.request.environ['SSL_CLIENT_S_DN_CN'] = 'doc'

        self.middleware(self.request)

        self.assertFalse(self.request.user.is_authenticated)

    def test_process_request_with_enabled_and_no_username(self):
        """Testing x509_auth_middleware with backend enabled and
        no username environment variable present
        """
        with self.siteconfig_settings(self.enabled_settings):
            self.middleware(self.request)

        self.assertFalse(self.request.user.is_authenticated)

    def test_process_request_with_enabled_and_username(self):
        """Testing x509_auth_middleware with backend enabled and
        username environment variable present
        """
        self.request.environ['SSL_CLIENT_S_DN_CN'] = 'doc'

        with self.siteconfig_settings(self.enabled_settings):
            self.middleware(self.request)

        self.assertTrue(self.request.user.is_authenticated)
        self.assertEqual(self.request.user.username, 'doc')



"""Review Board version and package information.

These variables and functions can be used to identify the version of
Review Board. They're largely used for packaging purposes.
"""

from __future__ import annotations

from reviewboard._version import (
    VERSION,
    __version__,
    __version_info__,
    get_package_version,
    get_version_string,
    is_release,
)


def get_manual_url():
    """Return the URL to the Review Board manual for this version."""
    if VERSION[2] == 0 and VERSION[4] != 'final':
        manual_ver = 'dev'
    else:
        manual_ver = '%s.%s' % (VERSION[0], VERSION[1])

    return 'https://www.reviewboard.org/docs/manual/%s/' % manual_ver


def initialize(load_extensions=True,
               setup_logging=True,
               setup_templates=True):
    """Begin initialization of Review Board.

    This sets up the logging, generates cache serial numbers, loads extensions,
    and sets up other aspects of Review Board. Once it has finished, it will
    fire the :py:data:`reviewboard.signals.initializing` signal.

    This must be called at some point before most features will work, but it
    will be called automatically in a standard install. If you are writing
    an extension or management command, you do not need to call this yourself.

    Args:
        load_extensions (bool, optional):
            Whether extensions should be automatically loaded upon
            initialization. If set, extensions will only load if the site
            has been upgraded to the latest version of Review Board.

        setup_logging (bool, optional):
            Whether to set up logging based on the configured settings.
            This can be disabled if the caller has their own logging
            configuration.

        setup_templates (bool, optional):
            Whether to set up state for template rendering. This can be
            disabled if the caller has no need for template rendering of
            any kind. This does not prevent template rendering from
            happening, but may change the output of some templates.

            Keep in mind that many pieces of functionality, such as avatars
            and some management commands, may be impacted by this setting.
    """
    import logging
    import os

    os.environ.setdefault(str('DJANGO_SETTINGS_MODULE'),
                          str('reviewboard.settings'))

    import settings_local

    # Set RBSITE_PYTHON_PATH to the path we need for any RB-bundled
    # scripts we may call.
    os.environ[str('RBSITE_PYTHONPATH')] = \
        os.path.dirname(settings_local.__file__)

    from django import setup
    from django.apps import apps

    if not apps.ready:
        setup()

    from django.conf import settings
    from django.db import DatabaseError
    from djblets import log
    from djblets.cache.serials import generate_ajax_serial
    from djblets.siteconfig.models import SiteConfiguration

    from reviewboard import signals
    from reviewboard.admin.siteconfig import load_site_config
    from reviewboard.extensions.base import get_extension_manager

    is_running_test = getattr(settings, 'RUNNING_TEST', False)

    if setup_logging and not is_running_test:
        # Set up logging.
        log.init_logging()

    load_site_config()

    if (setup_templates or load_extensions) and not is_running_test:
        if settings.DEBUG:
            logging.debug("Log file for Review Board v%s (PID %s)" %
                          (get_version_string(), os.getpid()))

        # Generate the AJAX serial, used for AJAX request caching.
        generate_ajax_serial()

        # Store the AJAX serial as a template serial, so we have a reference
        # to the real serial last modified timestamp of our templates. This
        # is useful since the extension manager will be modifying AJAX_SERIAL
        # to prevent stale caches for templates using hooks. Not all templates
        # use hooks, and may want to base cache keys off TEMPLATE_SERIAL
        # instead.
        #
        # We only want to do this once, so we don't end up replacing it
        # later with a modified AJAX_SERIAL later.
        if not getattr(settings, 'TEMPLATE_SERIAL', None):
            settings.TEMPLATE_SERIAL = settings.AJAX_SERIAL

    siteconfig = SiteConfiguration.objects.get_current()

    if load_extensions and not is_running_test:
        installed_version = get_version_string()

        if siteconfig.version == installed_version:
            # Load all extensions
            try:
                get_extension_manager().load()
            except DatabaseError:
                # This database is from a time before extensions, so don't
                # attempt to load any extensions yet.
                pass
        else:
            logging.warning('Extensions will not be loaded. The site must '
                            'be upgraded from Review Board %s to %s.',
                            siteconfig.version, installed_version)

    signals.initializing.send(sender=None)


def finalize_setup(is_upgrade=False):
    """Internal function to upgrade internal state after installs/upgrades.

    This should only be called by Review Board install or upgrade code.

    Version Changed:
        5.0:
            Removed the ``register_scmtools`` argument.

    Version Added:
        4.0:

    Args:
        is_upgrade (bool, optional):
            Whether this is finalizing an upgrade, rather than a new install.
    """
    from reviewboard import signals
    from reviewboard.admin.management.sites import init_siteconfig

    # Update the recorded product version.
    init_siteconfig()

    # Notify anything else that needs to listen.
    signals.finalized_setup.send(sender=None,
                                 is_upgrade=is_upgrade)


__all__ = [
    'VERSION',
    '__version__',
    '__version_info__',
    'finalize_setup',
    'get_manual_url',
    'get_package_version',
    'get_version_string',
    'initialize',
    'is_release',
]



"""Base support for writing hosting services.

.. autosummary::
   :nosignatures:

   ~reviewboard.hostingsvcs.base.bug_tracker.BaseBugTracker
   ~reviewboard.hostingsvcs.base.client.HostingServiceClient
   ~reviewboard.hostingsvcs.base.hosting_service.BaseHostingService
   ~reviewboard.hostingsvcs.base.http.HostingServiceHTTPRequest
   ~reviewboard.hostingsvcs.base.http.HostingServiceHTTPResponse
   ~reviewboard.hostingsvcs.base.registry.hosting_service_registry
   ~reviewboard.hostingsvcs.base.repository.RemoteRepository

Version Added:
    6.0
"""

from reviewboard.hostingsvcs.base.bug_tracker import BaseBugTracker
from reviewboard.hostingsvcs.base.client import HostingServiceClient
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.hostingsvcs.base.http import (HostingServiceHTTPRequest,
                                               HostingServiceHTTPResponse)
from reviewboard.hostingsvcs.base.registry import hosting_service_registry
from reviewboard.hostingsvcs.base.repository import RemoteRepository


__all__ = [
    'BaseBugTracker',
    'BaseHostingService',
    'HostingServiceClient',
    'HostingServiceHTTPRequest',
    'HostingServiceHTTPResponse',
    'RemoteRepository',
    'hosting_service_registry',
]

__autodoc_excludes__ = __all__



"""Code safety checker for detecting "Trojan Source" attacks.

Version Added:
    5.0
"""

from __future__ import annotations

import logging
import unicodedata
from itertools import chain
from typing import Tuple, TYPE_CHECKING

from django.utils.html import format_html
from django.utils.safestring import SafeString, mark_safe
from django.utils.translation import gettext, gettext_lazy as _
from typing_extensions import TypeAlias

from reviewboard.codesafety.checkers.base import (BaseCodeSafetyChecker,
                                                  CodeSafetyCheckResults,
                                                  CodeSafetyContentItem)

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Sequence
    from typing import ClassVar


logger = logging.getLogger(__name__)


_UnicodeRange: TypeAlias = Tuple[int, int]
_UnicodeRanges: TypeAlias = Tuple[_UnicodeRange, ...]


#: Zero-width Unicode characters.
#:
#: These characters can be used in some languages to make two different
#: identifiers appear to look the same.
#:
#: This regex covers the following characters:
#:
#: * ZWSP - Zero Width Space (U+200B)
#: * ZWNJ - Zero Width Non-Joiner (U+200C)
#:
#: Version Added:
#:     5.0
ZERO_WIDTH_UNICODE_CHAR_RANGES: _UnicodeRanges = (
    (0x200B, 0x200C),
)

#: Bi-directional Unicode characters.
#:
#: These can be used to alter the display of code, making a block of code
#: appear one way but actually execute another way.
#:
#: This regex covers the following characters:
#:
#: * LRE  - Left-to-Right Embedding (U+202A)
#: * RLE  - Right-to-Left Embedding (U+202B)
#: * LRO  - Left-to-Right Override (U+202D)
#: * RLO  - Right-to-Left Override (U+202E)
#: * LRI  - Left-to-Right Isolate (U+2066)
#: * RLI  - Right-to-Left Isolate (U+2067)
#: * FSI  - First Strong Isolate (U+2068)
#: * PDF  - Pop Directional Formatting (U+202C)
#: * PDI  - Pop Directional Isolate (U+2069)
#:
#: This is :cve:`2021-42574`.
#:
#: Version Added:
#:     5.0
BIDI_UNICODE_RANGES: _UnicodeRanges = (
    (0x202A, 0x202E),
    (0x2066, 0x2069),
)


class TrojanSourceCodeSafetyChecker(BaseCodeSafetyChecker):
    """Code safety checker for detecting "Trojan Source" attacks.

    Trojan Source attacks (:cve:`2021-42574`) enable a bad actor to write code
    that appears one way but behaves another way. This is done by taking
    advantage of Unicode bi-directional control characters or zero-width spaces
    to alter the way code is rendered, helping sneak malicious code past
    reviewers.

    This code safety checker looks for zero-width spaces and bi-directional
    control characters in lines. If found, warnings are flagged, characters
    are replaced with Unicode markers, and an alert is shown at the top of a
    diff, explaining the problem.

    See https://www.trojansource.codes/ for details on the attack.

    Version Added:
        5.0
    """

    checker_id: ClassVar[str] = 'trojan_source'
    summary = _('Possible "Trojan Source" code')

    file_alert_html_template_name = 'codesafety/trojan_source_alert.html'

    result_labels = {
        'bidi': _('Bi-directional Unicode characters (CVE-2021-42574)'),
        'confusable': _('Confusable Unicode characters (CVE-2021-42694)'),
        'zws': _('Zero-width space characters (CVE-2021-42574)'),
    }

    _unsafe_unicode_check_map: (Mapping[_UnicodeRange, str] | None) = None

    _check_unicode_ranges: Mapping[str, _UnicodeRanges] = {
        'bidi': BIDI_UNICODE_RANGES,
        'zws': ZERO_WIDTH_UNICODE_CHAR_RANGES,
    }

    @classmethod
    def get_main_confusable_aliases(
        cls,
    ) -> Sequence[str]:
        """Return a list of main Unicode aliases that can be customized.

        Returns:
            list of str:
            The list of aliases.
        """
        from reviewboard.codesafety._unicode_confusables import \
            CONFUSABLES_ID_TO_ALIAS_MAP

        return sorted(CONFUSABLES_ID_TO_ALIAS_MAP)

    def check_content(
        self,
        content_items: Sequence[CodeSafetyContentItem],
        *,
        check_confusables: bool = True,
        confusable_aliases_allowed: (Sequence[str] | None) = None,
        **kwargs,
    ) -> CodeSafetyCheckResults:
        """Check content for possible Trojan Source code.

        This will scan the characters of each line, looking for any
        bi-directional or zero-width space Unicode characters. If found,
        appropriate warnings will be returned.

        Args:
            content_items (list of dict):
                A list of dictionaries containing files and lines to check.

            check_confusables (bool, optional):
                Whether to check the line for Unicode confusables.

            confusable_aliases_allowed (list of str, optional):
                A list of Unicode aliases to exclude from Unicode confusables
                checks.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            dict:
            Results from the checks, which may contain the following:

            Keys:
                warnings (list of str):
                    A list of Unicode warning IDs, local to this safety
                    checker.
        """
        if confusable_aliases_allowed is None:
            confusable_aliases_allowed = []

        num_possible_warnings = len(self.result_labels)
        warnings = set()

        lines = (
            _line
            for _item in content_items
            for _line in _item['lines']
        )

        # Iterate over all characters across the lines. Since we're only
        # looking for Unicode characters, and we'll probably encounter a lot
        # of simple ASCII characters, don't bother with anything under ASCII
        # 128.
        chars = (
            _char
            for _char in chain.from_iterable(lines)
            if ord(_char) >= 128
        )

        unsafe_chars_iter = self._iter_unsafe_chars(
            chars,
            check_confusables=check_confusables,
            confusable_aliases_allowed=confusable_aliases_allowed)

        for _i, _c, _codepoint, check_name in unsafe_chars_iter:
            warnings.add(check_name)

            if len(warnings) == num_possible_warnings:
                # We don't need to check any more characters. We've
                # found at least one instance of everything we can
                # report.
                break

        return {
            'warnings': warnings,
        }

    def update_line_html(
        self,
        line_html: str,
        result_ids: Sequence[str],
        *,
        check_confusables: bool = True,
        confusable_aliases_allowed: (Sequence[str] | None) = None,
        **kwargs,
    ) -> SafeString:
        """Update the rendered diff HTML for a line.

        This will highlight any Unicode characters that would have triggered
        warnings, displaying the Unicode character code instead of rendering
        the character itself.

        Args:
            line_html (str):
                The HTML of the line.

            result_ids (list of str, unused):
                The list of result IDs that were found for the line.

            check_confusables (bool, optional):
                Whether to check the line for Unicode confusables.

            confusable_aliases_allowed (list of str, optional):
                A list of Unicode aliases to exclude from Unicode confusables
                checks.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            django.utils.safestring.SafeString:
            The updated HTML.
        """
        if confusable_aliases_allowed is None:
            confusable_aliases_allowed = []

        result = []
        last_append_i = 0

        unsafe_chars_iter = self._iter_unsafe_chars(
            line_html,
            check_confusables=check_confusables,
            confusable_aliases_allowed=confusable_aliases_allowed)

        for i, c, codepoint, _check_name in unsafe_chars_iter:
            result.append(line_html[last_append_i:i])

            # If it looks like a DUC and quacs like a DUC, then it's a
            # Displayed Unicode Character.
            #
            # We'll convert the character to a hex value and set the codepoint
            # and character entity on the tag so that the CSS can toggle its
            # display.
            try:
                char_name = unicodedata.name(c).title()
            except ValueError:
                char_name = gettext('Unknown')

            result.append(format_html(
                '<span class="rb-o-duc" data-codepoint="{codepoint}"'
                ' data-char="&#x{codepoint};" title="{title}"></span>',
                codepoint=f'{codepoint:X}',
                title=(
                    _('Unicode Character: %s')
                    % char_name
                )))
            last_append_i = i + 1

        result.append(line_html[last_append_i:])

        return mark_safe(''.join(result))

    def _iter_unsafe_chars(
        self,
        chars: Iterable[str],
        *,
        check_confusables: bool,
        confusable_aliases_allowed: Sequence[str],
    ) -> Iterator[tuple[int, str, int, str]]:
        """Iterate through a string, yielding unsafe characters.

        Args:
            chars (str):
                The characters to iterate through.

            check_confusables (bool, optional):
                Whether to check the line for Unicode confusables.

            confusable_aliases_allowed (list of str, optional):
                A list of Unicode aliases to exclude from Unicode confusables
                checks.

        Yields:
            tuple:
            Information on an unsafe character. This contains:

            Tuple:
                0 (int):
                    The 0-based index of the character in the provided string.

                1 (str):
                    The Unicode character.

                2 (int):
                    The Unicode codepoint.

                3 (str):
                    The result ID.
        """
        # We're importing this here, rather than at the module level, since
        # we want to avoid taking the hit until we need it the first time.
        from reviewboard.codesafety._unicode_confusables import (
            COMMON_CONFUSABLES_MAP,
            CONFUSABLES_ALIAS_TO_ID_MAP,
            ConfusablesMap,
        )

        confusables_map: ConfusablesMap = {}
        confusables_lang_ids_allowed: set[int] = set()

        if check_confusables:
            confusables_map = COMMON_CONFUSABLES_MAP

            # Check if any specific languages have been opted into. These will
            # be excluded from any confusable checks.
            confusables_lang_ids_allowed = {
                CONFUSABLES_ALIAS_TO_ID_MAP[_alias]
                for _alias in confusable_aliases_allowed
                if _alias in CONFUSABLES_ALIAS_TO_ID_MAP
            }

        checks_map = self._get_unsafe_unicode_check_map()

        for i, c in enumerate(chars):
            codepoint = ord(c)

            if c in confusables_map:
                if (not confusables_lang_ids_allowed or
                    confusables_map[c][1] not in confusables_lang_ids_allowed):
                    yield i, c, codepoint, 'confusable'
            else:
                for check_range, check_name in checks_map.items():
                    if check_range[0] <= codepoint <= check_range[1]:
                        yield i, c, codepoint, check_name
                        break

    @classmethod
    def _get_unsafe_unicode_check_map(cls) -> Mapping[_UnicodeRange, str]:
        """Return a range check map for matching unsafe Unicode characters.

        This is cached for all future instances.

        Returns:
            dict:
            The resulting range check map.
        """
        checks_map = cls._unsafe_unicode_check_map

        if checks_map is None:
            checks_map = {
                _range: _check_id
                for _check_id, _ranges in cls._check_unicode_ranges.items()
                for _range in _ranges
            }
            cls._unsafe_unicode_check_map = checks_map

        return checks_map



"""Utilities for creating FileDiffs."""

from __future__ import annotations

import os
from copy import deepcopy
from typing import (Iterator, Mapping, Optional, Protocol, Sequence,
                    TYPE_CHECKING, Union)

from django.utils.encoding import force_bytes, force_str
from django.utils.translation import gettext as _
from housekeeping import deprecate_non_keyword_only_args
from typing_extensions import TypedDict

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.diffviewer.errors import EmptyDiffError
from reviewboard.scmtools.core import (FileLookupContext,
                                       PRE_CREATION,
                                       Revision,
                                       UNKNOWN)
from reviewboard.scmtools.errors import FileNotFoundError

if TYPE_CHECKING:
    from django.http import HttpRequest

    from reviewboard.diffviewer.models import DiffCommit, DiffSet, FileDiff
    from reviewboard.diffviewer.parser import (BaseDiffParser,
                                               ParsedDiff,
                                               ParsedDiffFile)
    from reviewboard.scmtools.core import SCMTool
    from reviewboard.scmtools.models import Repository

    class _GetFileExistsFunc(Protocol):
        def __call__(
            self,
            *,
            path: str,
            revision: str,
            context: FileLookupContext,
        ) -> bool:
            ...


class _PreparedDiffInfo(TypedDict):
    """Intermediary information on a prepared diff.

    This is used for the FileDiff preparation stages. This is available
    only for internal typing.

    Version Added:
        8.0
    """

    #: All parsed files in the diff.
    files: Sequence[ParsedDiffFile]

    #: A mapping of modified filenames to parsed diff files.
    #:
    #: Each filename corresponds to a file in :py:attr:`files`.
    parent_files: Mapping[bytes, ParsedDiffFile]

    #: The parsed diff file.
    parsed_diff: ParsedDiff

    #: The parsed diff file for the parent diff.
    parsed_parent_diff: Optional[ParsedDiff]

    #: The parent diff file.
    parser: BaseDiffParser


@deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
def create_filediffs(
    *,
    diff_file_contents: bytes,
    parent_diff_file_contents: Optional[bytes],
    repository: Repository,
    basedir: Optional[Union[bytes, str]],
    base_commit_id: Optional[str],
    diffset: DiffSet,
    request: Optional[HttpRequest] = None,
    check_existence: bool = True,
    get_file_exists: Optional[_GetFileExistsFunc] = None,
    diffcommit: Optional[DiffCommit] = None,
    validate_only: bool = False,
) -> Sequence[FileDiff]:
    """Create FileDiffs from the given data.

    Version Changed:
        8.0:
        All arguments are now keyword-only arguments. Passing as positional
        arguments is deprecated and will be removed in Review Board 9.

    Args:
        diff_file_contents (bytes):
            The contents of the diff file.

        parent_diff_file_contents (bytes):
            The contents of the parent diff file.

        repository (reviewboard.scmtools.models.Repository):
            The repository the diff is being posted against.

        basedir (bytes or str):
            The base directory to prepend to all file paths in the diff.

        base_commit_id (str):
            The ID of the commit that the diff is based upon. This is
            needed by some SCMs or hosting services to properly look up
            files, if the diffs represent blob IDs instead of commit IDs
            and the service doesn't support those lookups.

        diffset (reviewboard.diffviewer.models.diffset.DiffSet):
            The DiffSet to attach the created FileDiffs to.

        request (django.http.HttpRequest, optional):
            The current HTTP request.

        check_existence (bool, optional):
            Whether or not existence checks should be performed against
            the upstream repository.

            This argument defaults to ``True``.

        get_file_exists (callable, optional):
            A callable that is used to determine if a file exists.

            This must be provided if ``check_existence`` is ``True``.

        diffcommit (reviewboard.diffviewer.models.diffcommit.DiffCommit,
                    optional):
            The DiffCommit to attach the created FileDiffs to.

        validate_only (bool, optional):
            Whether to just validate and not save. If ``True``, then this
            won't populate the database at all and will return ``None``
            upon success. This defaults to ``False``.

    Returns:
        list of reviewboard.diffviewer.models.filediff.FileDiff:
        The created FileDiffs.

        If ``validate_only`` is ``True``, the returned list will be empty.
    """
    from reviewboard.diffviewer.diffutils import convert_to_unicode
    from reviewboard.diffviewer.models import FileDiff

    diff_info = _prepare_diff_info(
        diff_file_contents=diff_file_contents,
        parent_diff_file_contents=parent_diff_file_contents,
        repository=repository,
        request=request,
        basedir=force_bytes(basedir),
        check_existence=check_existence,
        get_file_exists=get_file_exists,
        base_commit_id=base_commit_id)

    parent_files = diff_info['parent_files']
    parsed_diff = diff_info['parsed_diff']
    parsed_parent_diff = diff_info['parsed_parent_diff']
    parser = diff_info['parser']

    encoding_list = repository.get_encoding_list()

    # Copy over any extra_data for the DiffSet and DiffCommit, if any were
    # set by the parser.
    #
    # We'll do this even if we're validating, to ensure the data can be
    # copied over fine.
    main_extra_data = deepcopy(parsed_diff.extra_data)
    change_extra_data = deepcopy(parsed_diff.changes[0].extra_data)

    if change_extra_data:
        if diffcommit is not None:
            # We've already checked in _parse_diff that there's only a single
            # change in the diff, so we can assume that here.
            diffcommit.extra_data.update(change_extra_data)
        else:
            main_extra_data['change_extra_data'] = change_extra_data

    if main_extra_data:
        diffset.extra_data.update(main_extra_data)

    if parsed_parent_diff is not None:
        parent_extra_data = deepcopy(parsed_parent_diff.extra_data)
        parent_change_extra_data = deepcopy(
            parsed_parent_diff.changes[0].extra_data)

        if parent_change_extra_data:
            if diffcommit is not None:
                diffcommit.extra_data['parent_extra_data'] = \
                    parent_change_extra_data
            else:
                parent_extra_data['change_extra_data'] = \
                    parent_change_extra_data

        if parent_extra_data:
            diffset.extra_data['parent_extra_data'] = parent_extra_data

    # Convert the list of parsed files into FileDiffs.
    filediffs: list[FileDiff] = []

    for f in diff_info['files']:
        parent_file: Optional[ParsedDiffFile] = None
        parent_content: bytes = b''

        extra_data = f.extra_data.copy()

        if parsed_parent_diff is not None:
            if f.orig_filename:
                parent_file = parent_files.get(f.orig_filename)

            if parent_file is not None:
                parent_content = parent_file.data

                # Store the information on the parent's filename and revision.
                # It's important we force these to text, since they may be
                # byte strings and the revision may be a Revision instance.
                parent_source_filename = parent_file.orig_filename
                parent_source_revision = parent_file.orig_file_details

                parent_is_empty = (
                    parent_file.insert_count == 0 and
                    parent_file.delete_count == 0
                )

                if parent_file.moved or parent_file.copied:
                    extra_data['parent_moved'] = True

                if parent_file.extra_data:
                    extra_data['parent_extra_data'] = \
                        parent_file.extra_data.copy()
            else:
                # We don't have an entry, but we still want to record the
                # parent ID, so we have something in common for all the files
                # when looking up the source revision to fetch from the
                # repository.
                parent_is_empty = True
                parent_source_filename = f.orig_filename
                parent_source_revision = f.orig_file_details

                if (parent_source_revision != PRE_CREATION and
                    parsed_diff.uses_commit_ids_as_revisions):
                    # Since the file wasn't explicitly provided in the parent
                    # diff, but the ParsedDiff says that commit IDs are used
                    # as revisions, we can use its parent commit ID as the
                    # parent revision here.
                    parent_commit_id = \
                        parsed_parent_diff.changes[0].parent_commit_id
                    assert parent_commit_id

                    parent_source_revision = parent_commit_id

            # If this is a Revision, we'll need to ensure we specifically
            # cast it to a bytes before we convert it.
            if isinstance(parent_source_revision, Revision):
                parent_source_revision = bytes(parent_source_revision)

            # Store the information on the parent's filename and revision.
            # It's important we force these to text, since they may be
            # byte strings and the revision may be a Revision instance.
            #
            # Starting in Review Board 4.0.5, we store this any time there's
            # a parent diff, whether or not the file existed in the parent
            # diff.
            extra_data.update({
                FileDiff._IS_PARENT_EMPTY_KEY: parent_is_empty,
                'parent_source_filename':
                    convert_to_unicode(parent_source_filename,
                                       encoding_list)[1],
                'parent_source_revision':
                    convert_to_unicode(parent_source_revision,
                                       encoding_list)[1],
            })

        orig_file = convert_to_unicode(f.orig_filename, encoding_list)[1]
        dest_file = convert_to_unicode(f.modified_filename, encoding_list)[1]

        if f.deleted:
            status = FileDiff.DELETED
        elif f.moved:
            status = FileDiff.MOVED
        elif f.copied:
            status = FileDiff.COPIED
        else:
            status = FileDiff.MODIFIED

        filediff = FileDiff(
            diffset=diffset,
            commit=diffcommit,
            source_file=parser.normalize_diff_filename(orig_file),
            dest_file=parser.normalize_diff_filename(dest_file),
            source_revision=force_str(f.orig_file_details),
            dest_detail=force_str(f.modified_file_details),
            binary=f.binary,
            status=status,
            extra_data=extra_data)

        # Set this unconditionally, for backwards-compatibility purposes.
        # Review Board 4.0.6 introduced attribute wrappers in FileDiff and
        # introduced symlink targets. We ideally would not set this unless
        # it's True, but we don't want to risk breaking any assumptions on
        # its presence at this time.
        filediff.is_symlink = f.is_symlink

        if f.is_symlink:
            if f.old_symlink_target:
                filediff.old_symlink_target = \
                    convert_to_unicode(f.old_symlink_target, encoding_list)[1]

            if f.new_symlink_target:
                filediff.new_symlink_target = \
                    convert_to_unicode(f.new_symlink_target, encoding_list)[1]

        filediff.old_unix_mode = f.old_unix_mode
        filediff.new_unix_mode = f.new_unix_mode

        if not validate_only:
            # This state all requires making modifications to the database.
            # We only want to do this if we're saving.
            filediff.diff = f.data
            filediff.parent_diff = parent_content

            filediff.set_line_counts(raw_insert_count=f.insert_count,
                                     raw_delete_count=f.delete_count)

        filediffs.append(filediff)

    if not validate_only:
        FileDiff.objects.bulk_create(filediffs)

        if diffset.extra_data:
            diffset.save(update_fields=('extra_data',))

        if diffcommit is not None and diffcommit.extra_data:
            diffcommit.save(update_fields=('extra_data',))

    return filediffs


def _prepare_diff_info(
    *,
    diff_file_contents: bytes,
    parent_diff_file_contents: Optional[bytes],
    repository: Repository,
    request: Optional[HttpRequest],
    basedir: bytes,
    check_existence: bool,
    get_file_exists: Optional[_GetFileExistsFunc] = None,
    base_commit_id: Optional[str] = None,
) -> _PreparedDiffInfo:
    """Extract information and files from a diff.

    Version Changed:
        8.0:
        All arguments are now keyword-only arguments.

    Args:
        diff_file_contents (bytes):
            The contents of the diff.

        parent_diff_file_contents (bytes):
            The contents of the parent diff, if any.

        repository (reviewboard.scmtools.models.Repository):
            The repository against which the diff was created.

        request (django.http.HttpRequest):
            The current HTTP request.

        basedir (bytes):
            The base directory to prepend to all file paths in the diff.

        check_existence (bool):
            Whether or not existence checks should be performed against
            the upstream repository.

        get_file_exists (callable, optional):
            A callable to use to determine if a file exists in the repository.

            This argument must be provided if ``check_existence`` is ``True``.

        base_commit_id (unicode, optional):
            The ID of the commit that the diff is based upon. This is
            needed by some SCMs or hosting services to properly look up
            files, if the diffs represent blob IDs instead of commit IDs
            and the service doesn't support those lookups.

    Returns:
        _PreparedDiffInfo:
        A dictionary of information about the diff and parser.

    Raises:
        reviewboard.diffviewer.errors.EmptyDiffError:
            The diff contains no files.

        ValueError:
            ``check_existence`` was ``True`` but ``get_file_exists`` was not
            provided.
    """
    if check_existence and get_file_exists is None:
        raise ValueError('Must provide get_file_exists when check_existence '
                         'is True')

    tool = repository.get_scmtool()
    parsed_diff = _parse_diff(tool=tool,
                              diff_content=diff_file_contents)

    files = list(_process_files(
        parsed_diff=parsed_diff,
        basedir=basedir,
        repository=repository,
        base_commit_id=base_commit_id,
        request=request,
        check_existence=(check_existence and
                         not parent_diff_file_contents),
        get_file_exists=get_file_exists))

    if len(files) == 0:
        raise EmptyDiffError(_('The diff is empty.'))

    parsed_parent_diff = None
    parent_files: dict[bytes, ParsedDiffFile] = {}

    if parent_diff_file_contents:
        diff_filenames = {
            f.orig_filename
            for f in files
            if f.orig_filename
        }
        parsed_parent_diff = _parse_diff(
            tool=tool,
            diff_content=parent_diff_file_contents)

        # If the user supplied a base diff, we need to parse it and later
        # apply each of the files that are in main diff.
        parent_files = {
            f.modified_filename: f
            for f in _process_files(
                get_file_exists=get_file_exists,
                parsed_diff=parsed_parent_diff,
                basedir=basedir,
                repository=repository,
                base_commit_id=base_commit_id,
                request=request,
                check_existence=check_existence,
                limit_to=diff_filenames)
            if f.modified_filename
        }

    return {
        'files': files,
        'parent_files': parent_files,
        'parsed_diff': parsed_diff,
        'parsed_parent_diff': parsed_parent_diff,
        'parser': parsed_diff.parser,
    }


def _parse_diff(
    *,
    tool: SCMTool,
    diff_content: bytes,
) -> ParsedDiff:
    """Parse a diff using the SCMTool's diff parser.

    Version Changed:
        8.0:
        All arguments are now keyword-only arguments.

    Args:
        tool (reviewboard.scmtools.core.SCMTool):
            The tool providing the diff parser.

        diff_content (bytes):
            The diff content to parse.

    Returns:
        reviewboard.diffviewer.parser.ParsedDiff:
        The resulting parsed diff file.
    """
    parser = tool.get_parser(diff_content)
    parsed_diff = parser.parse_diff()

    # We can only support parsing a single change at this time, and we have to
    # have exactly one. Future architectural work will be needed to allow a
    # single diff to be uploaded that spans multiple commits.
    assert len(parsed_diff.changes) == 1, (
        '%s.parse_diff() must extract exactly one change from a diff.'
        % type(parser).__name__)

    return parsed_diff


def _process_files(
    *,
    parsed_diff: ParsedDiff,
    basedir: bytes,
    repository: Repository,
    base_commit_id: Optional[str],
    request: Optional[HttpRequest],
    get_file_exists: Optional[_GetFileExistsFunc] = None,
    check_existence: bool = False,
    limit_to: Optional[set[bytes]] = None,
) -> Iterator[ParsedDiffFile]:
    """Collect metadata about files in the parser.

    Version Changed:
        8.0:
        All arguments are now keyword-only arguments.

    Args:
        parsed_diff (reviewboard.diffviewer.parser.ParsedDiff):
            The parsed diff to process.

        basedir (bytes):
            The base directory to prepend to all file paths in the diff.

        repository (reviewboard.scmtools.models.Repository):
            The repository that the diff was created against.

        base_commit_id (str):
            The ID of the commit that the diff is based upon. This is
            needed by some SCMs or hosting services to properly look up
            files, if the diffs represent blob IDs instead of commit IDs
            and the service doesn't support those lookups.

        request (django.http.HttpRequest):
            The current HTTP request.

        check_existence (bool, optional):
            Whether or not existence checks should be performed against
            the upstream repository.

        get_file_exists (callable, optional):
            A callable to use to determine if a given file exists in the
            repository.

            If ``check_existence`` is ``True`` this argument must be
            provided.

        limit_to (set of bytes, optional):
            A list of filenames to limit the results to.

    Yields:
       reviewboard.diffviewer.parser.ParsedDiffFile:
       Each file present in the diff.

    Raises:
        ValueError:
            ``check_existence`` was ``True`` but ``get_file_exists`` was not
            provided.
    """
    if check_existence and get_file_exists is None:
        raise ValueError('Must provide get_file_exists when check_existence '
                         'is True')

    tool = repository.get_scmtool()
    parsed_change = parsed_diff.changes[0]

    for f in parsed_change.files:
        # This will either be a Revision or bytes. Either way, convert it
        # bytes now.
        orig_revision = force_bytes(f.orig_file_details)

        assert f.orig_filename is not None
        assert f.modified_filename is not None

        source_filename, source_revision = tool.parse_diff_revision(
            f.orig_filename,
            orig_revision,
            moved=f.moved,
            copied=f.copied)

        assert isinstance(source_filename, bytes), (
            '%s.parse_diff_revision() must return a bytes filename, not %r'
            % (type(tool).__name__, type(source_filename)))
        assert isinstance(source_revision, (bytes, Revision)), (
            '%s.parse_diff_revision() must return a revision which is either '
            'bytes or reviewboard.scmtools.core.Revision, not %r'
            % (type(tool).__name__, type(source_revision)))

        dest_filename = _normalize_filename(filename=f.modified_filename,
                                            basedir=basedir)

        if limit_to is not None and dest_filename not in limit_to:
            # This file isn't actually needed for the diff, so save
            # ourselves a remote file existence check and some storage.
            continue

        source_filename = _normalize_filename(filename=source_filename,
                                              basedir=basedir)

        if (check_existence and
            source_revision not in (PRE_CREATION, UNKNOWN) and
            not f.binary and
            not f.deleted and
            not f.moved and
            not f.copied):
            assert get_file_exists is not None

            context = FileLookupContext(
                request=request,
                base_commit_id=base_commit_id,
                diff_extra_data=parsed_diff.extra_data,
                commit_extra_data=parsed_change.extra_data,
                file_extra_data=f.extra_data)

            if not get_file_exists(path=force_str(source_filename),
                                   revision=force_str(source_revision),
                                   context=context):
                raise FileNotFoundError(path=force_str(source_filename),
                                        revision=force_str(source_revision),
                                        base_commit_id=base_commit_id,
                                        context=context)

            # If validation found a more suitable source revision, then set
            # that instead of what was parsed out of the diff. This is
            # important for SCMs like Mercurial, which support multiple
            # commits but don't have per-file revision information in diffs,
            # so we don't even know the commit that introduced the last change
            # to a file. We can only resolve this during the validation
            # phase.
            #
            # This was added in Review Board 7.0.2.
            validated_parent_id = \
                f.extra_data.pop('__validated_parent_id', None)

            if validated_parent_id is not None:
                assert isinstance(validated_parent_id, str)

                source_revision = validated_parent_id.encode('utf-8')

        f.orig_filename = source_filename
        f.orig_file_details = source_revision
        f.modified_filename = dest_filename

        yield f


def _normalize_filename(
    *,
    filename: bytes,
    basedir: bytes,
) -> bytes:
    """Normalize a filename to be relative to the repository root.

    Version Changed:
        8.0:
        All arguments are now keyword-only arguments.

    Args:
        filename (bytes):
            The filename to normalize.

        basedir (bytes):
            The base directory to prepend to all file paths in the diff.

    Returns:
        bytes:
        The filename relative to the repository root.
    """
    assert isinstance(filename, bytes)
    assert isinstance(basedir, bytes)

    if filename.startswith(b'/'):
        return filename

    return os.path.join(basedir, filename).replace(b'\\', b'/')



"""Datagrids for the Dashboard and other pages."""

from __future__ import annotations

import pytz
from typing import List, Optional, TYPE_CHECKING

from django.contrib.auth.models import User
from django.db.models import Q
from django.http import Http404
from django.utils.translation import gettext_lazy as _
from djblets.datagrid.grids import (
    Column,
    DateTimeColumn,
    DataGrid as DjbletsDataGrid,
    AlphanumericDataGrid as DjbletsAlphanumericDataGrid)
from djblets.util.templatetags.djblets_utils import ageid

from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.datagrids.columns import (BugsColumn,
                                           DateTimeSinceColumn,
                                           DiffSizeColumn,
                                           DiffUpdatedColumn,
                                           DiffUpdatedSinceColumn,
                                           FullNameColumn,
                                           GroupMemberCountColumn,
                                           GroupsColumn,
                                           MyCommentsColumn,
                                           NewUpdatesColumn,
                                           PendingCountColumn,
                                           PeopleColumn,
                                           RepositoryColumn,
                                           ReviewCountColumn,
                                           ReviewGroupStarColumn,
                                           ReviewRequestCheckboxColumn,
                                           ReviewRequestIDColumn,
                                           ReviewRequestStarColumn,
                                           ReviewSummaryColumn,
                                           ShipItColumn,
                                           SummaryColumn,
                                           ToMeColumn,
                                           UsernameColumn)
from reviewboard.datagrids.sidebar import Sidebar, DataGridSidebarMixin
from reviewboard.datagrids.builtin_items import (IncomingSection,
                                                 OutgoingSection,
                                                 OverviewSection,
                                                 UserGroupsItem,
                                                 UserProfileItem)
from reviewboard.reviews.models import Group, ReviewRequest, Review
from reviewboard.site.models import LocalSite
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from django.http import HttpRequest
    from typelets.django.strings import StrOrPromise

    from reviewboard.accounts.models import Profile


class ShowClosedReviewRequestsMixin(object):
    """A mixin for showing or hiding closed review requests."""

    def load_extra_state(self, profile, allow_hide_closed=True):
        """Load extra state for the datagrid."""
        if profile:
            self.show_closed = profile.show_closed

        try:
            self.show_closed = (
                int(self.request.GET.get('show-closed',
                                         self.show_closed)) != 0)
        except ValueError:
            # do nothing
            pass

        if allow_hide_closed and not self.show_closed:
            self.queryset = self.queryset.filter(**{
                self.status_query_field: 'P',
            })

        self.queryset = self.queryset.filter(**{
            self.site_query_field: self.local_site,
        })

        if profile and self.show_closed != profile.show_closed:
            profile.show_closed = self.show_closed
            return ['show_closed']

        return []


class DataGridJSMixin(object):
    """Mixin that provides enhanced JavaScript support for datagrids.

    This contains additional information on the JavaScript views/models
    to load for the page, allowing for enhanced functionality in datagrids.
    """

    #: A list of extra CSS static bundles to load on the page.
    css_bundle_names = []

    #: A list of extra JavaScript static bundles to load on the page.
    js_bundle_names = []

    #: The JavaScript Model to use for the page state.
    js_model_class = 'RB.DatagridPage'

    #: The JavaScript View to use for the page rendering.
    js_view_class = 'RB.DatagridPageView'

    #: Whether or not to periodically reload the contents of the datagrid.
    periodic_reload = False

    #: Extra data to pass to the JavaScript Model.
    extra_js_model_data = None

    def __init__(self, *args, **kwargs):
        """Initialize the mixin.

        This will pull out the Local Site, which is common to all datagrids,
        and store it for later use and for JavaScript attribute population.

        Args:
            *args (tuple):
                Positional arguments passed to the datagrid.

            **kwargs (dict):
                Keyword arguments passed to the datagrid.
        """
        self.local_site = kwargs.pop('local_site', None)

        super(DataGridJSMixin, self).__init__(*args, **kwargs)

    def get_js_model_attrs(self):
        """Return attributes for the JavaScript model.

        These will be passed to the model specified in
        :py:attr:`js_model_class` during construction.

        Subclasses can override this to provide additional data.

        Returns:
            dict:
            Attributes to provide to the JavaScript model.
        """
        attrs = {}

        if self.extra_js_model_data:
            attrs['data'] = self.extra_js_model_data

        if self.local_site:
            attrs['localSiteName'] = self.local_site.name

        return attrs

    def get_js_model_options(self):
        """Return options for the JavaScript model.

        These will be passed to the model specified in
        :py:attr:`js_model_class` during construction.

        Subclasses can override this to provide additional data.

        Returns:
            dict:
            Options to provide to the JavaScript model.
        """
        return {}

    def get_js_view_options(self):
        """Return options for the JavaScript view.

        These will be passed to the view specified in
        :py:attr:`js_view_class` during construction.

        Subclasses can override this to provide additional data.

        Returns:
            dict:
            Options to provide to the JavaScript view.
        """
        return {
            'periodicReload': self.periodic_reload,
        }


class DataGrid(DataGridJSMixin, DjbletsDataGrid):
    """Base class for a datagrid in Review Board.

    This contains additional information on JavaScript views/models
    to load for the page.
    """


class AlphanumericDataGrid(DataGridJSMixin, DjbletsAlphanumericDataGrid):
    """Base class for an alphanumeric datagrid in Review Board.

    This contains additional information on JavaScript views/models
    to load for the page.
    """


class ReviewRequestDataGrid(ShowClosedReviewRequestsMixin, DataGrid):
    """A datagrid that displays a list of review requests.

    This datagrid accepts the show_closed parameter in the URL, allowing
    submitted review requests to be filtered out or displayed.
    """

    new_updates = NewUpdatesColumn()
    my_comments = MyCommentsColumn()
    star = ReviewRequestStarColumn()
    ship_it = ShipItColumn()
    summary = SummaryColumn()
    submitter = UsernameColumn(label=_('Owner'),
                               user_relation=['submitter'])

    branch = Column(
        label=_('Branch'),
        db_field='branch',
        shrink=True,
        sortable=True,
        link=False)
    bugs_closed = BugsColumn()
    repository = RepositoryColumn()
    time_added = DateTimeColumn(
        label=_('Posted'),
        detailed_label=_('Posted Time'),
        format='F jS, Y, P',
        shrink=True,
        css_class=lambda r: ageid(r.time_added))
    last_updated = DateTimeColumn(
        label=_('Last Updated'),
        format='F jS, Y, P',
        shrink=True,
        db_field='last_updated',
        field_name='last_updated',
        css_class=lambda r: ageid(r.last_updated))
    diff_updated = DiffUpdatedColumn(
        format='F jS, Y, P',
        shrink=True,
        css_class=lambda r: ageid(r.diffset_history.last_diff_updated))
    time_added_since = DateTimeSinceColumn(
        label=_('Posted'),
        detailed_label=_('Posted Time (Relative)'),
        field_name='time_added', shrink=True,
        css_class=lambda r: ageid(r.time_added))
    last_updated_since = DateTimeSinceColumn(
        label=_('Last Updated'),
        detailed_label=_('Last Updated (Relative)'), shrink=True,
        db_field='last_updated',
        field_name='last_updated',
        css_class=lambda r: ageid(r.last_updated))
    diff_updated_since = DiffUpdatedSinceColumn(
        detailed_label=_('Diff Updated (Relative)'),
        shrink=True,
        css_class=lambda r: ageid(r.diffset_history.last_diff_updated))
    diff_size = DiffSizeColumn()

    review_count = ReviewCountColumn()

    target_groups = GroupsColumn()
    target_people = PeopleColumn()
    to_me = ToMeColumn()

    review_id = ReviewRequestIDColumn()

    status_query_field = 'status'
    site_query_field = 'local_site'

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the datagrid.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent.

            **kwargs (dict):
                Keyword arguments to pass to the parent.
        """
        super().__init__(*args, **kwargs)

        self.listview_template = 'datagrids/review_request_listview.html'
        self.profile_sort_field = 'sort_review_request_columns'
        self.profile_columns_field = 'review_request_columns'
        self.show_closed = True
        self.submitter_url_name = 'user'
        self.default_sort = ['-last_updated']
        self.default_columns = [
            'star',
            'new_updates',
            'ship_it',
            'summary',
            'submitter',
            'time_added',
            'last_updated_since',
        ]

        # Add local timezone info to the columns.
        user = self.request.user

        if user.is_authenticated:
            profile = user.get_profile()
            self.timezone = pytz.timezone(profile.timezone)
            self.time_added.timezone = self.timezone
            self.last_updated.timezone = self.timezone
            self.diff_updated.timezone = self.timezone

    def load_extra_state(self, profile, allow_hide_closed=True):
        """Load extra state for the datagrid."""
        return super(ReviewRequestDataGrid, self).load_extra_state(
            profile, allow_hide_closed)

    def post_process_queryset(
        self,
        queryset: QuerySet[ReviewRequest],
    ) -> QuerySet[ReviewRequest]:
        """Add additional data to the queryset.

        Args:
            queryset (django.db.models.QuerySet):
                The queryset to post-process.

        Returns:
            django.db.models.QuerySet:
            The resulting queryset.
        """
        queryset = queryset.with_counts(self.request.user)

        if self.local_site is not None:
            # Make sure to include the Local Site if one is specified, so
            # that accessing it on the review request on each row won't
            # result in extra database queries.
            queryset = queryset.select_related('local_site')

        return super().post_process_queryset(queryset)

    def link_to_object(self, state, obj, value):
        """Return a link to the given object."""
        if value and isinstance(value, User):
            return local_site_reverse('user', request=self.request,
                                      args=[value])

        return obj.get_absolute_url()


class ReviewDataGrid(ShowClosedReviewRequestsMixin, DataGrid):
    """A datagrid that displays a list of reviews.

    This datagrid accepts the show_closed parameter in the URL, allowing
    submitted review requests to be filtered out or displayed.
    """

    timestamp = DateTimeColumn(
        label=_('Date Reviewed'),
        format='F jS, Y',
        shrink=True)
    submitter = UsernameColumn(label=_('Owner'),
                               user_relation=['review_request', 'submitter'])
    review_summary = ReviewSummaryColumn()

    status_query_field = 'review_request__status'
    site_query_field = 'review_request__local_site'

    def __init__(self, *args, **kwargs):
        """Initialize the datagrid."""
        super(ReviewDataGrid, self).__init__(*args, **kwargs)

        self.listview_template = 'datagrids/review_request_listview.html'
        self.show_closed = True
        self.default_sort = ['-timestamp']
        self.default_columns = [
           'submitter', 'review_summary', 'timestamp',
        ]

        # Add local timezone info to the columns.
        user = self.request.user

        if user.is_authenticated:
            profile = user.get_profile()
            self.timezone = pytz.timezone(profile.timezone)
            self.timestamp.timezone = self.timezone

    def post_process_queryset(
        self,
        queryset: QuerySet[Review],
    ) -> QuerySet[Review]:
        """Add additional data to the queryset.

        Args:
            queryset (django.db.models.QuerySet):
                The queryset to post-process.

        Returns:
            django.db.models.QuerySet:
            The resulting queryset.
        """
        if self.local_site is not None:
            # Make sure to include the Local Site if one is specified, so
            # that accessing it on the review request on each row won't
            # result in extra database queries.
            queryset = queryset.select_related('review_request__local_site')

        return super().post_process_queryset(queryset)


class DashboardDataGrid(DataGridSidebarMixin, ReviewRequestDataGrid):
    """Displays the dashboard.

    The dashboard is the main place where users see what review requests
    are out there that may need their attention.
    """

    new_updates = NewUpdatesColumn()
    my_comments = MyCommentsColumn()
    selected = ReviewRequestCheckboxColumn()

    sidebar = Sidebar(
        [
            OverviewSection,
            OutgoingSection,
            IncomingSection,
        ],
        default_view_id='overview')

    js_model_class = 'RB.Dashboard'
    js_view_class = 'RB.DashboardView'
    periodic_reload = True

    def __init__(self, *args, **kwargs):
        """Initialize the datagrid."""
        super(DashboardDataGrid, self).__init__(*args, **kwargs)

        self.listview_template = 'datagrids/hideable_listview.html'
        self.profile_sort_field = 'sort_dashboard_columns'
        self.profile_columns_field = 'dashboard_columns'
        self.default_view = 'overview'
        self.show_closed = False
        self.show_archived = False
        self.default_sort = ['-last_updated']
        self.default_columns = [
            'selected', 'new_updates', 'ship_it', 'my_comments', 'summary',
            'submitter', 'last_updated_since'
        ]

        self.extra_js_model_data = {
            'show_archived': self.show_archived,
        }

        self.user = self.request.user
        self.profile = self.user.get_profile()
        self.site_profile = self.user.get_site_profile(self.local_site)

    def load_extra_state(
        self,
        profile: Profile,
    ) -> List[str]:
        """Load extra state for the datagrid.

        Args:
            profile (reviewboard.accounts.models.Profile):
                The profile containing state for customizing the dashboard.

        Returns:
            list of str:
            A list of profile fields that have changed.
        """
        request = self.request
        group_name = request.GET.get('group', '')
        view = request.GET.get('view', self.default_view)
        user = request.user
        changed_fields = []

        queryset: QuerySet[ReviewRequest]
        title: StrOrPromise

        if view == 'outgoing':
            queryset = ReviewRequest.objects.from_user(
                user,  # The target user
                user,  # The accessing user
                distinct=False,
                local_site=self.local_site)
            title = _('All Outgoing Review Requests')
        elif view == 'overview':
            queryset = ReviewRequest.objects.to_or_from_user(
                user,  # The target user
                user,  # The accessing user
                distinct=False,
                local_site=self.local_site)
            title = _('Open Incoming and Outgoing Review Requests')
        elif view == 'mine':
            queryset = ReviewRequest.objects.from_user(
                user,  # The target user
                user,  # The accessing user
                status=None,
                distinct=False,
                local_site=self.local_site)
            title = _('All My Review Requests')
        elif view == 'to-me':
            queryset = ReviewRequest.objects.to_user_directly(
                user,  # The target user
                user,  # The accessing user
                distinct=False,
                local_site=self.local_site)
            title = _('Incoming Review Requests to Me')
        elif view in ('to-group', 'to-watched-group'):
            if group_name:
                # to-group is special because we want to make sure that the
                # group exists and show a 404 if it doesn't. Otherwise, we'll
                # show an empty datagrid with the name.
                try:
                    group = (
                        Group.objects
                        .get(Q(name=group_name) &
                             LocalSite.objects.build_q(self.local_site,
                                                       allow_all=False))
                    )

                    if not group.is_accessible_by(user):
                        raise Http404
                except Group.DoesNotExist:
                    raise Http404

                queryset = ReviewRequest.objects.to_group(
                    group_name=group_name,
                    local_site=self.local_site,
                    distinct=False,
                    user=user)
                title = _('Incoming Review Requests to %s') % group_name
            else:
                queryset = ReviewRequest.objects.to_user_groups(
                    username=user,  # The target user
                    user=user,      # The accessing user
                    distinct=False,
                    local_site=self.local_site)
                title = _('All Incoming Review Requests to My Groups')
        elif view == 'starred':
            queryset = self.profile.starred_review_requests.public(
                user=user,
                local_site=self.local_site,
                distinct=False,
                status=None)
            title = _('Starred Review Requests')
        elif view == 'incoming':
            queryset = ReviewRequest.objects.to_user(
                user,  # The target user
                user,  # The accessing user
                distinct=False,
                local_site=self.local_site)
            title = _('All Incoming Review Requests')
        else:
            raise Http404

        if profile and 'show_archived' in profile.extra_data:
            show_archived = profile.extra_data['show_archived']
        else:
            show_archived = self.show_archived

        try:
            show = request.GET.get('show-archived', show_archived)
            show_archived = (int(show) != 0)
        except ValueError:
            pass

        if not show_archived:
            # This may produce a large number of archived review requests.
            # Rather than work with a large number of IDs, we'll use a
            # subquery here.
            queryset = queryset.exclude(pk__in=(
                ReviewRequestVisit.objects
                .filter(user=user)
                .exclude(visibility=ReviewRequestVisit.VISIBLE)
                .values_list('review_request_id', flat=True)
            ))

        if (profile and
            show_archived != profile.extra_data.get('show_archived')):
            profile.extra_data['show_archived'] = show_archived
            changed_fields.append('extra_data')

        self.extra_js_model_data['show_archived'] = show_archived
        self.show_archived = show_archived
        self.queryset = queryset
        self.title = title

        changed_fields += super().load_extra_state(profile,
                                                   allow_hide_closed=False)

        return changed_fields


class UsersDataGrid(AlphanumericDataGrid):
    """A datagrid showing a list of users registered on Review Board."""

    username = UsernameColumn(label=_('Username'))
    fullname = FullNameColumn(label=_('Full Name'), link=True, expand=True)
    pending_count = PendingCountColumn(_('Open Review Requests'),
                                       field_name='directed_review_requests',
                                       shrink=True)

    def __init__(
        self,
        request: HttpRequest,
        queryset: QuerySet[User] = User.objects.all(),
        title: StrOrPromise = _('All users'),
        local_site: Optional[LocalSite] = None,
    ) -> None:
        """Initialize the datagrid.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            queryset (django.db.models.QuerySet, optional):
                The base queryset for this datagrid.

            title (str, optional):
                The title for this datagrid.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site being accessed.
        """
        # Note that if local_site is None, we want to show all users on the
        # server, so we can't unconditionally make this part of the query.
        #
        # If we have a Local Site instance, we have no reason to check if
        # Local Sites are available on the server. So we don't need to use
        # LocalSite.objects.build_q() or has_local_sites() here.
        if local_site:
            queryset = queryset.filter(local_site=local_site)

        super().__init__(
            request=request,
            queryset=queryset,
            title=title,
            sortable_column='username',
            extra_regex=r'^[0-9_\-\.].*')

        self.listview_template = 'datagrids/user_listview.html'
        self.default_sort = ['username']
        self.profile_sort_field = 'sort_submitter_columns'
        self.profile_columns_field = 'submitter_columns'
        self.default_columns = [
            'username', 'fullname', 'pending_count'
        ]
        self.show_inactive = False

    def link_to_object(self, state, obj, value):
        """Return a link to the given object."""
        return local_site_reverse('user', request=self.request,
                                  args=[obj.username])

    def load_extra_state(self, profile):
        """Load extra state for the datagrid.

        This handles hiding or showing inactive users.

        Args:
            profile (reviewboard.accounts.models.Profile):
                The user profile which contains some basic
                configurable settings.

        Returns:
            bool:
            Always returns False.
        """
        show_inactive = self.request.GET.get('show-inactive', 0)

        try:
            self.show_inactive = int(show_inactive)
        except ValueError:
            pass

        if not self.show_inactive:
            self.queryset = self.queryset.filter(is_active=True)

        return []

    def post_process_queryset_for_data(
        self,
        queryset: QuerySet,
        **kwargs,
    ) -> QuerySet:
        """Add additional lookups for the datagrid's queryset

        This will add any column-specific queries to the queryset, and then
        select-related the Local Sites so that operations that require the
        Local Site relation can be conducted without an additional query.

        Version Added:
            5.0.7

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

            **kwargs (dict):
                Additional keyword arguments for future expansion.

        Returns:
            django.db.models.query.QuerySet:
            The resulting augmented QuerySet.
        """
        return (
            super().post_process_queryset_for_data(queryset, **kwargs)
            .select_related('profile')
        )


class GroupDataGrid(DataGrid):
    """A datagrid showing a list of review groups accessible by the user."""

    star = ReviewGroupStarColumn()
    name = Column(_('Group ID'), link=True, sortable=True)
    displayname = Column(_('Group Name'), field_name='display_name',
                         link=True, expand=True)
    pending_count = PendingCountColumn(_('Open Review Requests'),
                                       field_name='review_requests',
                                       link=True,
                                       shrink=True)
    member_count = GroupMemberCountColumn(_('Members'),
                                          field_name='members',
                                          shrink=True)

    def __init__(self, request, title=_('All groups'), *args, **kwargs):
        """Initialize the datagrid."""
        local_site = kwargs.get('local_site')

        # Build the queryset for the groups.
        queryset = Group.objects.accessible(request.user,
                                            local_site=local_site)

        super().__init__(request,
                         queryset=queryset,
                         title=title,
                         *args,
                         **kwargs)

        self.listview_template = 'datagrids/listview.html'
        self.profile_sort_field = 'sort_group_columns'
        self.profile_columns_field = 'group_columns'
        self.default_sort = ['name']
        self.default_columns = [
            'star', 'name', 'displayname', 'pending_count'
        ]

    def post_process_queryset_for_data(
        self,
        queryset: QuerySet,
        **kwargs,
    ) -> QuerySet:
        """Add additional lookups for the datagrid's queryset

        This will add any column-specific queries to the queryset, and then
        select-related the Local Sites so that operations that require the
        Local Site relation can be conducted without an additional query.

        Version Added:
            5.0.7

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

            **kwargs (dict):
                Additional keyword arguments for future expansion.

        Returns:
            django.db.models.query.QuerySet:
            The resulting augmented QuerySet.
        """
        queryset = super().post_process_queryset_for_data(queryset, **kwargs)

        if self.local_site is not None:
            queryset = queryset.select_related('local_site')

        return queryset

    @staticmethod
    def link_to_object(state, obj, value):
        """Return a link to the given object."""
        return obj.get_absolute_url()


class UserPageDataGridMixin(DataGridSidebarMixin):
    """An abstract class for data grids on the user page.

    This will display information about the user on the side.
    """

    sidebar = Sidebar([
        UserProfileItem,
        UserGroupsItem,
    ])


class UserPageReviewRequestDataGrid(UserPageDataGridMixin,
                                    ReviewRequestDataGrid):
    """A data grid for the user page.

    This will show the review requests the user has out for review.
    """

    allow_search_indexing = False
    tab_title = _('Review Requests')

    def __init__(self, request, user, *args, **kwargs):
        """Initialize the datagrid."""
        queryset = ReviewRequest.objects.from_user(
            user.username,
            user=request.user,
            distinct=False,
            status=None,
            local_site=kwargs.get('local_site'),
            filter_private=True,
            show_inactive=True)

        super(UserPageReviewRequestDataGrid, self).__init__(
            request,
            queryset=queryset,
            title=_("%s's Review Requests") % user.username,
            extra_context={
                'pii_safe_title': _("User's Review Requests"),
            },
            *args, **kwargs)

        self.groups = user.review_groups.accessible(request.user)
        self.user = user


class UserPageReviewsDataGrid(UserPageDataGridMixin, ReviewDataGrid):
    """A data grid for the user page.

    This will show reviews the user has made on other review requests.
    """

    allow_search_indexing = False
    tab_title = _('Reviews')

    def __init__(self, request, user, *args, **kwargs):
        """Initialize the datagrid."""
        queryset = Review.objects.from_user(
            user.username,
            user=request.user,
            public=True,
            filter_private=True,
            status=None,
            local_site=kwargs.get('local_site'))

        super(UserPageReviewsDataGrid, self).__init__(
            request,
            queryset=queryset,
            title=_("%s's Reviews") % user.username,
            extra_context={
                'pii_safe_title': _("User's Reviews"),
            },
            *args, **kwargs)

        self.groups = user.review_groups.accessible(request.user)
        self.user = user



"""Views for batch operations.

Version Added:
    6.0
"""

import json
import logging
from enum import Enum
from typing import List

from django.contrib.auth.models import User
from django.db.models import Q
from django.http import (HttpRequest,
                         HttpResponse,
                         HttpResponseServerError,
                         JsonResponse)
from django.views.generic import View
from djblets.siteconfig.models import SiteConfiguration
from djblets.views.generic.base import CheckRequestMethodViewMixin

from reviewboard.accounts.mixins import CheckLoginRequiredViewMixin
from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.notifications.email.message import \
    prepare_batch_review_request_mail
from reviewboard.notifications.email.utils import send_email
from reviewboard.reviews.models import ReviewRequest, Review
from reviewboard.site.mixins import CheckLocalSiteAccessViewMixin


logger = logging.getLogger(__name__)


class BatchOperation(str, Enum):
    """A batch operation to run.

    Version Added:
        6.0
    """

    #: Publish a set of review requests, reviews, and replies.
    PUBLISH = 'publish'

    #: Close review requests as discarded.
    DISCARD = 'discard'

    #: Close review requests as submitted.
    CLOSE = 'close'

    #: Mark review requests as archived.
    ARCHIVE = 'archive'

    #: Mark review requests as muted.
    MUTE = 'mute'

    #: Mark review requests as visible (unarchive and/or unmute).
    UNARCHIVE = 'unarchive'


class BatchOperationView(CheckRequestMethodViewMixin,
                         CheckLoginRequiredViewMixin,
                         CheckLocalSiteAccessViewMixin,
                         View):
    """Performs batch operations.

    This is effectively an RPC endpoint that allows clients to initiate various
    batch operations.

    This currently includes:

    * Publishing multiple items, including review requests, reviews, and review
      replies. This is for the unified banner when publishing both review
      request updates and reviews all at once (and with a single e-mail
      notification).
    * Closing review requests, either as submitted or discarded.
    * Adjusting the visibility state of review requests (archive or mute).

    This is not considered a public API, and may change from version to
    version. No stability or compatibility guarantees are made.

    Version Added:
        6.0
    """

    def post(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle a POST request for this view.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple, unused):
                Positional arguments passed to the handler.

            **kwargs (dict, unused):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response to send to the client.
        """
        data = request.POST
        batch_str = data.get('batch')
        user = request.user
        assert isinstance(user, User)

        if not batch_str:
            return JsonResponse(
                data={
                    'stat': 'fail',
                    'error': 'Batch data was not found.',
                },
                status=400)

        try:
            batch_data = json.loads(batch_str)
        except Exception as e:
            logger.warning('Could not parse batch data "%s": %s',
                           batch_str, e,
                           extra={'request': request})
            return JsonResponse(
                data={
                    'stat': 'fail',
                    'error': 'Could not parse batch data: %s.' % e,
                },
                status=400)

        op_str = batch_data.get('op')

        try:
            op = BatchOperation(op_str)
        except ValueError:
            return JsonResponse(
                data={
                    'stat': 'fail',
                    'error': 'Unknown batch operation "%s".' % op_str,
                },
                status=400)

        review_request_ids = batch_data.get('review_requests')

        if review_request_ids:
            if not (isinstance(review_request_ids, list) and
                    all(isinstance(id, int) for id in review_request_ids)):
                return JsonResponse(
                    data={
                        'stat': 'fail',
                        'error': 'review_requests must be an array of integers.',
                    },
                    status=400)

            if self.local_site:
                id_field = 'local_id'
                extra_query = Q(local_id__in=review_request_ids)
            else:
                id_field = 'pk'
                extra_query = Q(pk__in=review_request_ids)

            review_requests_qs = (
                ReviewRequest.objects.public(
                    user=user,
                    status=None,
                    show_all_unpublished=True,
                    local_site=self.local_site,
                    extra_query=extra_query)
            )

            if op == BatchOperation.PUBLISH:
                review_requests_qs = review_requests_qs.select_related('draft')

            # We use in_bulk to deduplicate review requests. It's possible that
            # the .public() query can return duplicates, and we don't want to
            # use .distinct() because of performance issues with MySQL. Ideally
            # we'd just be able to avoid using extra_query and pass in the
            # id_field to this like we do for reviews below, but local_id isn't
            # unique (even though in the context of the resulting queryset it
            # is).
            review_requests = review_requests_qs.in_bulk()

            if len(review_requests) != len(review_request_ids):
                found_ids = [
                    getattr(rr, id_field)
                    for rr in review_requests.values()
                ]
                missing_ids = sorted(set(review_request_ids) - set(found_ids))

                if self.local_site:
                    return JsonResponse(
                        data={
                            'stat': 'fail',
                            'error': ('The following review requests are not '
                                      'valid for the local site: %r.'
                                      % missing_ids),
                        },
                        status=400)
                else:
                    return JsonResponse(
                        data={
                            'stat': 'fail',
                            'error': ('The following review requests are not '
                                      'valid: %r.'
                                      % missing_ids),
                        },
                        status=400)

            ordered_review_requests = [
                item[1]
                for item in sorted(review_requests.items())
            ]
        else:
            ordered_review_requests = []

        if op == BatchOperation.PUBLISH:
            review_ids = batch_data.get('reviews', [])
            trivial = batch_data.get('trivial', False)
            archive_after_publish = batch_data.get('archive', False)

            if review_ids:
                if not (isinstance(review_ids, list) and
                        all(isinstance(id, int) for id in review_ids)):
                    return JsonResponse(
                        data={
                            'stat': 'fail',
                            'error': 'reviews must be an array of integers.',
                        },
                        status=400)

                reviews_qs = (
                    Review.objects.accessible(
                        base_reply_to=Review.objects.ANY,
                        user=user,
                        local_site=self.local_site)
                    .prefetch_related('comments')
                    .select_related('user', 'base_reply_to'))

                # We use in_bulk to deduplicate reviews. It's possible that the
                # .accessible() query can return duplicates, and we don't want
                # to use .distinct() because of performance issues with MySQL.
                reviews = reviews_qs.in_bulk(review_ids)

                if len(reviews) != len(review_ids):
                    found_ids = reviews.keys()
                    missing_ids = sorted(set(review_ids) - set(found_ids))

                    if self.local_site:
                        return JsonResponse(
                            data={
                                'stat': 'fail',
                                'error': ('The following reviews are not valid '
                                          'for the local site: %r.'
                                          % missing_ids),
                            },
                            status=400)
                    else:
                        return JsonResponse(
                            data={
                                'stat': 'fail',
                                'error': (
                                    'The following reviews are not valid: %r.'
                                    % missing_ids),
                            },
                            status=400)
                review_objects = list(reviews.values())
            else:
                review_objects = []

            return self._publish(
                request=request,
                user=user,
                review_requests=ordered_review_requests,
                reviews=review_objects,
                trivial=trivial,
                archive_after_publish=archive_after_publish)
        elif op == BatchOperation.CLOSE:
            return self._close(
                request=request,
                user=user,
                review_requests=ordered_review_requests,
                close_type=ReviewRequest.SUBMITTED)
        elif op == BatchOperation.DISCARD:
            return self._close(
                request=request,
                user=user,
                review_requests=ordered_review_requests,
                close_type=ReviewRequest.DISCARDED)
        elif op == BatchOperation.ARCHIVE:
            return self._set_visibility(
                request=request,
                user=user,
                review_requests=ordered_review_requests,
                visibility=ReviewRequestVisit.ARCHIVED)
        elif op == BatchOperation.MUTE:
            return self._set_visibility(
                request=request,
                user=user,
                review_requests=ordered_review_requests,
                visibility=ReviewRequestVisit.MUTED)
        elif op == BatchOperation.UNARCHIVE:
            # This also means unmute.
            return self._set_visibility(
                request=request,
                user=user,
                review_requests=ordered_review_requests,
                visibility=ReviewRequestVisit.VISIBLE)
        else:
            logger.error('Hit unknown batch operation case: %s',
                         op_str,
                         extra={'request': request})
            return HttpResponseServerError()

    def _publish(
        self,
        *,
        request: HttpRequest,
        user: User,
        review_requests: List[ReviewRequest],
        reviews: List[Review],
        trivial: bool,
        archive_after_publish: bool,
    ) -> JsonResponse:
        """Publish a set of review requests and reviews.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            user (django.contrib.auth.models.User):
                The user initiating the publish operation.

            review_requests (list of reviewboard.reviews.models.ReviewRequest):
                The review requests to publish.

            reviews (list of reviewboard.reviews.models.Review):
                The reviews (and review replies) to publish.

            trivial (bool):
                Whether to skip sending e-mail notifications.

            archive_after_publish (bool):
                Whether to archive affected review requests after publishing.

        Returns:
            django.http.JsonResponse:
            The response to send to the client.
        """
        for review_request in review_requests:
            if not review_request.is_mutable_by(user):
                return JsonResponse(
                    data={
                        'stat': 'fail',
                        'error': ('User does not have permission to publish '
                                  'review request %d.'
                                  % review_request.display_id),
                    },
                    status=403)
            elif not review_request.get_draft():
                return JsonResponse(
                    data={
                        'stat': 'fail',
                        'error': ('No draft found for review request %d.'
                                  % review_request.display_id),
                    },
                    status=404)

        for review in reviews:
            if not review.is_mutable_by(user):
                return JsonResponse(
                    data={
                        'stat': 'fail',
                        'error': ('User does not have permission to publish '
                                  'review %d.'
                                  % review.pk),
                    },
                    status=403)
            elif review.public:
                return JsonResponse(
                    data={
                        'stat': 'fail',
                        'error': ('Review %d is already published.'
                                  % review.pk),
                    },
                    status=404)

            can_publish, err = review.can_publish(
                review_request_will_publish=(review.review_request in
                                             review_requests))

            if not can_publish:
                return JsonResponse(
                    data={
                        'stat': 'fail',
                        'error': err,
                    },
                    status=403)

        review_requests_to_archive = set()

        # We publish with trivial=True to skip the individual notifications,
        # and then we'll send one batched notification at the end. However, in
        # the case where we're calling this with only a single thing to
        # publish, go through the normal notifications path.
        num_review_requests = len(review_requests)
        num_reviews = len(reviews)

        suppress_notifications = \
            trivial or (num_review_requests + len(reviews)) != 1
        batches = {}

        for i, review_request in enumerate(review_requests):
            try:
                changes = review_request.publish(
                    user=user,
                    trivial=suppress_notifications)

                batches[review_request.pk] = {
                    'review_request': review_request,
                    'review_request_changed': True,
                    'reviews': [],
                    'review_replies': [],
                    'changedesc': changes,
                }

                if archive_after_publish:
                    review_requests_to_archive.add(review_request)
            except Exception as e:
                logger.exception('Failed to publish review request %d: %s',
                                 review_request.pk, e,
                                 extra={'request': request})

                if i >= 1:
                    stat = 'mixed'
                else:
                    stat = 'fail'

                return JsonResponse(
                    data={
                        'stat': stat,
                        'error': ('Failed to publish review request %d: %s'
                                  % (review_request.display_id, e)),
                        'review_requests_published': i,
                        'review_requests_not_published':
                            num_review_requests - i,
                        'reviews_published': 0,
                        'reviews_not_published': num_reviews,
                    },
                    status=500)

        for i, review in enumerate(reviews):
            try:
                review.publish(
                    user=user,
                    trivial=suppress_notifications,
                    review_request_will_publish=(review.review_request in
                                                 review_requests))

                review_request_id = review.review_request.pk

                if review_request_id not in batches:
                    batches[review_request_id] = {
                        'review_request': review.review_request,
                        'review_request_changed': False,
                        'reviews': [],
                        'review_replies': [],
                        'changedesc': None,
                    }

                if review.base_reply_to:
                    batches[review_request_id]['review_replies'].append(review)
                else:
                    batches[review_request_id]['reviews'].append(review)

                if archive_after_publish:
                    review_requests_to_archive.add(review.review_request)
            except Exception as e:
                logger.exception('Failed to publish review %d: %s',
                                 review.pk, e, extra={'request': request})

                if i > 1 or num_review_requests > 0:
                    stat = 'mixed'
                else:
                    stat = 'fail'

                return JsonResponse(
                    data={
                        'stat': stat,
                        'error': ('Failed to publish review %d: %s'
                                  % (review.pk, e)),
                        'review_requests_published': num_review_requests,
                        'review_requests_not_published': 0,
                        'reviews_published': i,
                        'reviews_not_published': num_reviews - i,
                    },
                    status=500)

        # Now prepare and send the batched notifications, if appropriate. We do
        # one e-mail per review request.
        siteconfig = SiteConfiguration.objects.get_current()

        if (suppress_notifications and
            not trivial and
            siteconfig.get('mail_send_review_mail')):
            for batch in batches.values():
                send_email(prepare_batch_review_request_mail,
                           user=user,
                           request=request,
                           **batch)

        # Finally, handle the "archive after publish" option. This affects all
        # published review requests, as well as all review requests associated
        # with reviews.
        for review_request in review_requests_to_archive:
            ReviewRequestVisit.objects.update_visibility(
                review_request, user, ReviewRequestVisit.ARCHIVED)

        return JsonResponse({
            'stat': 'ok',
        })

    def _close(
        self,
        *,
        request: HttpRequest,
        user: User,
        review_requests: List[ReviewRequest],
        close_type: str,
    ) -> JsonResponse:
        """Close a set of review requests.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            user (django.contrib.auth.models.User):
                The user initiating the operation.

            review_requests (list of reviewboard.reviews.models.ReviewRequest):
                The review requests to close.

            close_type (str):
                The status code to use. This is one of
                :py:attr:`reviewboard.reviews.models.ReviewRequest.STATUSES`.

        Returns:
            django.http.JsonResponse:
            The response to send to the client.
        """
        for review_request in review_requests:
            if (not review_request.is_mutable_by(user) and
                not user.has_perm('reviews.can_change_status',
                                  review_request.local_site)):
                return JsonResponse(
                    data={
                        'stat': 'fail',
                        'error': ('User does not have permission to close '
                                  'review request %d.'
                                  % review_request.display_id),
                    },
                    status=403)

        for i, review_request in enumerate(review_requests):
            try:
                review_request.close(close_type=close_type, user=user)
            except Exception as e:
                logger.exception('Failed to close review request %d: %s',
                                 review_request.pk, e,
                                 extra={'request': request})

                if i >= 1:
                    stat = 'mixed'
                else:
                    stat = 'fail'

                return JsonResponse(
                    data={
                        'stat': stat,
                        'error': ('Failed to close review request %d: %s'
                                  % (review_request.display_id, e)),
                        'review_requests_closed': i,
                        'review_requests_not_closed': len(review_requests) - i,
                    },
                    status=500)

        return JsonResponse({
            'stat': 'ok',
            'review_requests_closed': len(review_requests),
            'review_requests_not_closed': 0,
        })

    def _set_visibility(
        self,
        *,
        request: HttpRequest,
        user: User,
        review_requests: List[ReviewRequest],
        visibility: str,
    ) -> JsonResponse:
        """Set the visibility of a set of review requests.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            user (django.contrib.auth.models.User):
                The user initiating the operation.

            review_requests (list of reviewboard.reviews.models.ReviewRequest):
                The review requests to set the visibility for.

            visibility (str):
                The visibility state. This is one of
                :py:attr:`reviewboard.accounts.models.ReviewRequestVisit.
                VISIBILITY`.

        Returns:
            django.http.JsonResponse:
            The response to send to the client.
        """
        for review_request in review_requests:
            ReviewRequestVisit.objects.update_visibility(
                review_request,
                user,
                visibility)

        return JsonResponse({
            'stat': 'ok',
        })



"""Administration form for file attachment storage settings."""

from django import forms
from django.utils.translation import gettext_lazy as _
from djblets.siteconfig.forms import SiteSettingsForm

from reviewboard.admin.checks import (get_can_use_amazon_s3,
                                      get_can_use_openstack_swift,
                                      get_can_use_couchdb)
from reviewboard.admin.siteconfig import load_site_config


class S3StorageSettingsForm(SiteSettingsForm):
    """Settings subform for S3-based file storage."""

    aws_access_key_id = forms.CharField(
        label=_('Amazon AWS access key'),
        help_text=_('Your Amazon AWS access key ID. This can be found in '
                    'the "Security Credentials" section of the AWS site.'),
        required=True,
        widget=forms.TextInput(attrs={'size': '40'}))

    aws_secret_access_key = forms.CharField(
        label=_('Amazon AWS secret access key'),
        help_text=_('Your Amazon AWS secret access ID. This can be found in '
                    'the "Security Credentials" section of the AWS site.'),
        required=True,
        widget=forms.TextInput(attrs={'size': '40'}))

    aws_s3_bucket_name = forms.CharField(
        label=_('S3 bucket name'),
        help_text=_('Bucket name inside Amazon S3.'),
        required=True,
        widget=forms.TextInput(attrs={'size': '40'}))

    aws_calling_format = forms.ChoiceField(
        label=_('Amazon AWS calling format'),
        choices=(
            (1, 'Path'),
            (2, 'Subdomain'),
            (3, 'Vanity'),
        ),
        help_text=_('Calling format for AWS requests.'),
        required=True)

    # TODO: these items are consumed in the S3Storage backend, but I'm not
    # totally sure what they mean, or how to let users set them via siteconfig
    # (especially AWS_HEADERS, which is a dictionary). For now, defaults will
    # suffice.
    #
    # 'aws_headers':            'AWS_HEADERS',
    # 'aws_default_acl':        'AWS_DEFAULT_ACL',
    # 'aws_querystring_active': 'AWS_QUERYSTRING_ACTIVE',
    # 'aws_querystring_expire': 'AWS_QUERYSTRING_EXPIRE',
    # 'aws_s3_secure_urls':     'AWS_S3_SECURE_URLS',

    def __init__(self, *args, **kwargs):
        """Initialize the subform.

        If Amazon S3 support isn't available, the form's fields will be
        disabled.

        Args:
            *args (tuple):
                Additional positional arguments for the parent class.

            **kwargs (dict):
                Additional keyword arguments for the parent class.
        """
        super(S3StorageSettingsForm, self).__init__(*args, **kwargs)

        can_use_amazon_s3, reason = get_can_use_amazon_s3()

        if not can_use_amazon_s3:
            self.disabled_fields['aws_access_key_id'] = True
            self.disabled_fields['aws_secret_access_key'] = True
            self.disabled_fields['aws_s3_bucket_name'] = True
            self.disabled_fields['aws_calling_format'] = True
            self.disabled_reasons['aws_access_key_id'] = reason

    class Meta:
        title = _('Amazon S3 Settings')
        fieldsets = (
            (None, {
                'classes': ('wide', 'hidden'),
                'fields': ('aws_access_key_id',
                           'aws_secret_access_key',
                           'aws_s3_bucket_name',
                           'aws_calling_format'),
            }),
         )


class CouchDBStorageSettingsForm(SiteSettingsForm):
    """Settings subform for CouchDB-based file storage.

    Note that this is currently unused. It's here for legacy reasons and
    future support.
    """

    couchdb_default_server = forms.CharField(
        label=_('Default server'),
        help_text=_('For example, "http://couchdb.local:5984"'),
        required=True)

    # TODO: this is consumed in the CouchDBStorage backend, but I'm not sure
    # how to let users set it via siteconfig, since it's a dictionary. Since I
    # haven't tested the CouchDB backend at all, it'll just sit here for now.
    #
    # 'couchdb_storage_options': 'COUCHDB_STORAGE_OPTIONS',

    def __init__(self, *args, **kwargs):
        """Initialize the subform.

        If CouchDB support isn't available, the form's fields will be
        disabled.

        Args:
            *args (tuple):
                Additional positional arguments for the parent class.

            **kwargs (dict):
                Additional keyword arguments for the parent class.
        """
        super(CouchDBStorageSettingsForm, self).__init__(*args, **kwargs)

        can_use_couchdb, reason = get_can_use_couchdb()

        if not can_use_couchdb:
            self.disabled_fields['couchdb_default_server'] = True
            self.disabled_reasons['couchdb_default_server'] = reason

    class Meta:
        title = _('CouchDB Settings')
        fieldsets = (
            (None, {
                'classes': ('wide', 'hidden'),
                'fields': ('couchdb_default_server',),
            }),
        )


class SwiftStorageSettingsForm(SiteSettingsForm):
    """Settings subform for OpenStack Swift-based file storage."""

    swift_auth_url = forms.CharField(
        label=_('Swift auth URL'),
        help_text=_('The URL for the auth server, '
                    'e.g. http://127.0.0.1:5000/v2.0'),
        required=True,
        widget=forms.TextInput(attrs={'size': '40'}))

    swift_username = forms.CharField(
        label=_('Swift username'),
        help_text=_('The username to use to authenticate, '
                    'e.g. system:root'),
        required=True,
        widget=forms.TextInput(attrs={'size': '40'}))

    swift_key = forms.CharField(
        label=_('Swift key'),
        help_text=_('The key (password) to use to authenticate.'),
        required=True,
        widget=forms.TextInput(attrs={'size': '40'}))

    swift_auth_version = forms.ChoiceField(
        label=_('Swift auth version'),
        choices=(
            ('1', _('1.0')),
            ('2', _('2.0')),
        ),
        help_text=_('The version of the authentication protocol to use.'),
        required=True)

    swift_container_name = forms.CharField(
        label=_('Swift container name'),
        help_text=_('The container in which to store the files. '
                    'This container must be publicly readable.'),
        required=True,
        widget=forms.TextInput(attrs={'size': '40'}))

    def __init__(self, *args, **kwargs):
        """Initialize the subform.

        If OpenStack Swift support isn't available, the form's fields will be
        disabled.

        Args:
            *args (tuple):
                Additional positional arguments for the parent class.

            **kwargs (dict):
                Additional keyword arguments for the parent class.
        """
        super(SwiftStorageSettingsForm, self).__init__(*args, **kwargs)

        can_use_openstack_swift, reason = get_can_use_openstack_swift()

        if not can_use_openstack_swift:
            self.disabled_fields['swift_auth_url'] = True
            self.disabled_fields['swift_username'] = True
            self.disabled_fields['swift_key'] = True
            self.disabled_fields['swift_auth_version'] = True
            self.disabled_fields['swift_container_name'] = True
            self.disabled_reasons['swift_auth_url'] = reason

    class Meta:
        title = _('OpenStack Swift Settings')
        fieldsets = (
            (None, {
                'classes': ('wide', 'hidden'),
                'fields': ('swift_auth_url',
                           'swift_username',
                           'swift_key',
                           'swift_auth_version',
                           'swift_container_name'),
            }),
        )


class StorageSettingsForm(SiteSettingsForm):
    """File storage backend settings for Review Board."""

    storage_backend = forms.ChoiceField(
        label=_('File storage method'),
        choices=(
            ('filesystem', _('Host file system')),
            ('s3', _('Amazon S3')),
            ('swift', _('OpenStack Swift')),
        ),
        help_text=_('Storage method and location for uploaded files, such as '
                    'screenshots and file attachments.'),
        required=True,
        widget=forms.Select(attrs={
            'data-subform-group': 'storage-backend',
        }))

    def __init__(self, *args, **kwargs):
        """Initialize the storage settings form.

        This will set up the list of available storage backend settings forms,
        allowing the browser to show the appropriate settings form based on the
        selected backend.

        Args:
            *args (tuple):
                Additional positional arguments for the parent class.

            **kwargs (dict):
                Additional keyword arguments for the parent class.
        """
        super(StorageSettingsForm, self).__init__(*args, **kwargs)

        self.storage_backend_forms = {
            's3': S3StorageSettingsForm(*args, **kwargs),
            'swift': SwiftStorageSettingsForm(*args, **kwargs),
        }

    def is_valid(self):
        """Return whether the form is valid.

        This will check the validity of the fields on this form and on the
        selected storage backend's settings form.

        Returns:
            bool:
            ``True`` if the main settings form and storage backend's settings
            form is valid. ``False`` if either form is invalid.
        """
        if not super(StorageSettingsForm, self).is_valid():
            return False

        backend_id = self.cleaned_data['storage_backend']
        backend_form = self.storage_backend_forms.get(backend_id)

        return backend_form is None or backend_form.is_valid()

    def save(self):
        """Save the form.

        This will write the new configuration to the database. It will then
        force a site configuration reload.
        """
        super(StorageSettingsForm, self).save()

        backend_id = self.cleaned_data['storage_backend']

        if backend_id in self.storage_backend_forms:
            backend_form = self.storage_backend_forms[backend_id]
            backend_form.save()

        load_site_config()

    class Meta:
        title = _('File Storage Settings')
        subforms = (
            {
                'subforms_attr': 'storage_backend_forms',
                'controller_field': 'storage_backend',
            },
        )



"""Unit tests for reviewboard.reviews.admin."""

from django.urls import reverse

from reviewboard.reviews.models import DefaultReviewer
from reviewboard.testing.testcase import TestCase


class DefaultReviewerFormTests(TestCase):
    """Tests for reviewboard.reviews.admin:DefaultReviewerAdmin."""

    fixtures = ['test_users']

    def test_defaultreviewer_form_redirect(self):
        """Testing that a DefaultReviewer form can render on page, and saves
        data correctly
        """
        self.assertTrue(self.client.login(username='admin', password='admin'))
        test_group = self.create_review_group()

        response = self.client.get(
            reverse('admin:reviews_defaultreviewer_add'))
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse('admin:reviews_defaultreviewer_add'),
            {
                'file_regex': '/',
                'groups': test_group.pk,
                'name': 'Test Group',
            })
        self.assertRedirects(
            response,
            reverse('admin:reviews_defaultreviewer_changelist'))
        default_reviewer = DefaultReviewer.objects.latest('pk')

        response = self.client.get(
            reverse('admin:reviews_defaultreviewer_change',
                    args=(default_reviewer.pk,)))



"""A hook for registering a hosting service."""

from __future__ import annotations

from django.utils.translation import gettext as _
from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint

from reviewboard.hostingsvcs.base import hosting_service_registry


class HostingServiceHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for registering a hosting service."""

    def initialize(self, service_cls):
        """Initialize the hook.

        This will register the hosting service.

        Args:
            service_cls (type):
                The hosting service class to register. This must be a
                subclass of :py:class:`~reviewboard.hostingsvcs.base.
                hosting_service.BaseHostingService`.

        Raises:
            ValueError:
                The service's :py:attr:`~reviewboard.hostingsvcs.base.
                .hosting_service.BaseHostingService.hosting_service_id`
                attribute was not set.
        """
        hosting_service_id = service_cls.hosting_service_id

        if hosting_service_id is None:
            raise ValueError(_('%s.hosting_service_id must be set.')
                             % (service_cls.__name__))

        self.hosting_service_id = hosting_service_id
        hosting_service_registry.register(service_cls)

    def shutdown(self):
        """Shut down the hook.

        This will unregister the hosting service.
        """
        hosting_service_registry.unregister_by_id(self.hosting_service_id)



"""Add ReviewRequest.issue_verifying_count.

Version Added:
    3.0
"""

from django_evolution.mutations import AddField
from djblets.db.fields import CounterField


MUTATIONS = [
    AddField('ReviewRequest', 'issue_verifying_count', CounterField,
             null=True),
]



from django import forms
from django.utils.translation import gettext_lazy as _

from reviewboard.hostingsvcs.base.forms import BaseHostingServiceRepositoryForm
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService


class FogBugzForm(BaseHostingServiceRepositoryForm):
    fogbugz_account_domain = forms.CharField(
        label=_('Account domain'),
        max_length=64,
        required=True,
        help_text=_('The domain used for your FogBugz site, as in '
                    'https://&lt;domain&gt;.fogbugz.com/'),
        widget=forms.TextInput(attrs={'size': '60'}))


class FogBugz(BaseHostingService):
    """Bug tracker support for FogBugz.

    FogBugz is a bug tracker service provided by Fog Creek. This integration
    supports linking bug numbers to reports on a FogBugz account.
    """
    name = _('FogBugz')
    hosting_service_id = 'fogbugz'
    supports_bug_trackers = True

    form = FogBugzForm
    bug_tracker_field = \
        'https://%(fogbugz_account_domain)s.fogbugz.com/f/cases/%%s'



"""Hooks for adding new review request fields."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint

from reviewboard.reviews.fields import (get_review_request_fieldset,
                                        register_review_request_fieldset,
                                        unregister_review_request_fieldset)


class ReviewRequestFieldSetsHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for creating fieldsets on the side of the review request page.

    A fieldset contains one or more fields, and is mainly used to separate
    groups of fields from each other.

    This takes a list of fieldset classes as parameters, which it will
    later instantiate as necessary. Each fieldset can be pre-populated with
    one or more custom field classes.
    """

    def initialize(self, fieldsets):
        """Initialize the hook.

        This will register each of the provided fieldsets for review
        requests.

        Args:
            fieldsets (list of type):
                The list of fieldset classes to register. Each must be a
                subclass of
                :py:class:`~reviewboard.reviews.fields.BaseReviewRequestFieldSet`.

        Raises:
            djblets.registries.errors.ItemLookupError:
                A fieldset was already registered matching an ID from this
                list.
        """
        self.fieldsets = fieldsets

        for fieldset in fieldsets:
            register_review_request_fieldset(fieldset)

    def shutdown(self):
        """Shut down the hook.

        This will unregister each of the fieldsets from the review requests.
        """
        for fieldset in self.fieldsets:
            unregister_review_request_fieldset(fieldset)


class ReviewRequestFieldsHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for creating fields on the review request page.

    This is used to create custom fields on a review request page for
    requesting and storing data. A field can be editable, or it can be only
    for display purposes. See the classes in
    :py:mod:`reviewboard.reviews.fields` for more information and
    documentation.

    This hook takes the ID of a registered fieldset where the provided
    field classes should be added. Review Board supplies three built-in
    fieldset IDs:

    ``main``:
        The fieldset with Description and Testing Done.

    ``info``:
        The "Information" fieldset on the side.

    ``reviewers``:
        The "Reviewers" fieldset on the side.

    Any registered fieldset ID can be provided, whether from this extension
    or another.

    Field classes can only be added to a single fieldset.
    """

    def initialize(self, fieldset_id, fields):
        """Initialize the hook.

        This will register each of the provided field classes into the
        fieldset with the given ID.

        Args:
            fieldset_id (unicode):
                The ID of the
                :py:class:`~reviewboard.reviews.fields.BaseReviewRequestFieldSet`
                to register.

            fields (list of type):
                The list of fields to register into the fieldset. Each must be
                a subclass of
                :py:class:`~reviewboard.reviews.fields.BaseReviewRequestField`.
        """
        self.fieldset_id = fieldset_id
        self.fields = fields

        fieldset = get_review_request_fieldset(fieldset_id)
        assert fieldset is not None

        for field_cls in fields:
            fieldset.add_field(field_cls)

    def shutdown(self):
        """Shut down the hook.

        This will unregister each of the field classes from the fieldset.
        """
        fieldset = get_review_request_fieldset(self.fieldset_id)
        assert fieldset is not None

        for field_cls in self.fields:
            fieldset.remove_field(field_cls)



"""Administration form for review workflow.

Version Added:
    8.0
"""

from __future__ import annotations

from django import forms
from django.utils.translation import gettext_lazy as _
from djblets.siteconfig.forms import SiteSettingsForm


class ReviewSettingsForm(SiteSettingsForm):
    """Review workflow settings for Review Board.

    Version Added:
        8.0
    """

    css_bundle_names = ['djblets-forms']
    js_bundle_names = ['djblets-forms']

    reviews_allow_self_shipit = forms.BooleanField(
        label=_('Allow users to mark "Ship It!" on their own review requests'),
        help_text=_(
            'If selected, users will be allowed to mark their own review '
            'requests as "Ship It!".'
        ),
        required=False,
    )

    class Meta:
        """Metadata for the form."""

        title = _('Review Workflow Settings')



"""Unit tests for reviewboard.notifications.forms.WebHookTargetForm."""

from reviewboard.notifications.forms import WebHookTargetForm
from reviewboard.notifications.models import WebHookTarget
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class WebHookTargetFormTests(TestCase):
    """Unit tests for reviewboard.notifications.forms.WebHookTargetForm."""

    fixtures = ['test_scmtools']

    def setUp(self) -> None:
        """Set up the test."""
        super().setUp()

        self.local_site = LocalSite.objects.create(name='test')

        # Create repositories for the test.
        self.local_site_repo = self.create_repository(
            name='local-site-repo',
            local_site=self.local_site)

        self.global_site_repo = self.create_repository(
            name='global-site-repo')

    def test_without_localsite(self) -> None:
        """Testing WebHookTargetForm without a LocalSite"""
        # Make sure the initial state and querysets are what we expect on init.
        form = WebHookTargetForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repositories'].queryset),
                         [self.global_site_repo, self.local_site_repo])

        # Now test what happens when it's been fed data and validated.
        form = WebHookTargetForm(data={
            'apply_to': WebHookTarget.APPLY_TO_SELECTED_REPOS,
            'encoding': WebHookTarget.ENCODING_JSON,
            'events': ['*'],
            'url': 'https://example.com/',
            'repositories': [self.global_site_repo.pk],
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repositories'].queryset),
                         [self.global_site_repo, self.local_site_repo])

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['repositories'].queryset),
                         [self.global_site_repo, self.local_site_repo])

        webhook = form.save()
        self.assertIsNone(webhook.local_site)
        self.assertEqual(list(webhook.repositories.all()),
                         [self.global_site_repo])

    def test_without_localsite_and_instance(self) -> None:
        """Testing WebHookTargetForm without a LocalSite and editing instance
        """
        webhook = WebHookTarget.objects.create()

        form = WebHookTargetForm(
            data={
                'apply_to': WebHookTarget.APPLY_TO_ALL,
                'encoding': WebHookTarget.ENCODING_JSON,
                'events': ['*'],
                'url': 'https://example.com/',
                'repositories': [self.global_site_repo.pk],
            },
            instance=webhook)
        self.assertTrue(form.is_valid())

        new_webhook = form.save()
        self.assertEqual(webhook.pk, new_webhook.pk)
        self.assertIsNone(new_webhook.local_site)

    def test_without_localsite_and_with_local_site_repo(self) -> None:
        """Testing WebHookTargetForm without a LocalSite and Repository on a
        LocalSite
        """
        form = WebHookTargetForm(data={
            'apply_to': WebHookTarget.APPLY_TO_ALL,
            'encoding': WebHookTarget.ENCODING_JSON,
            'events': ['*'],
            'url': 'https://example.com/',
            'repositories': [self.local_site_repo.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'repositories': ['Select a valid choice. 1 is not one of the '
                                 'available choices.'],
            })

    def test_with_limited_localsite(self) -> None:
        """Testing WebHookTargetForm limited to a LocalSite"""
        form = WebHookTargetForm(limit_to_local_site=self.local_site)

        self.assertEqual(form.limited_to_local_site, self.local_site)
        self.assertNotIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repositories'].queryset),
                         [self.local_site_repo])

    def test_with_limited_localsite_and_changing_site(self) -> None:
        """Testing WebHookTargetForm limited to a LocalSite and changing
        LocalSite
        """
        site2 = LocalSite.objects.create(name='test-site-2')

        form = WebHookTargetForm(
            data={
                'apply_to': WebHookTarget.APPLY_TO_ALL,
                'encoding': WebHookTarget.ENCODING_JSON,
                'events': ['*'],
                'url': 'https://example.com/',
                'repositories': [self.local_site_repo.pk],
                'local_site': site2.pk,
            },
            limit_to_local_site=self.local_site)

        self.assertEqual(form.limited_to_local_site, self.local_site)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['local_site'], self.local_site)

        webhook = form.save()
        self.assertEqual(webhook.local_site, self.local_site)

    def test_with_limited_localsite_and_compatible_instance(self) -> None:
        """Testing WebHookTargetForm limited to a LocalSite and editing
        compatible instance
        """
        webhook = WebHookTarget(local_site=self.local_site)

        # This should just simply not raise an exception.
        WebHookTargetForm(instance=webhook,
                          limit_to_local_site=self.local_site)

    def test_with_limited_localsite_and_incompatible_instance(self) -> None:
        """Testing WebHookTargetForm limited to a LocalSite and editing
        incompatible instance
        """
        webhook = WebHookTarget.objects.create()

        error_message = (
            'The provided instance is not associated with a LocalSite '
            'compatible with this form. Please contact support.'
        )

        with self.assertRaisesMessage(ValueError, error_message):
            WebHookTargetForm(instance=webhook,
                              limit_to_local_site=self.local_site)

    def test_with_limited_localsite_and_invalid_repository(self) -> None:
        """Testing WebHookTargetForm limited to a LocalSite with a Repository
        not on the LocalSite
        """
        form = WebHookTargetForm(
            data={
                'apply_to': WebHookTarget.APPLY_TO_ALL,
                'encoding': WebHookTarget.ENCODING_JSON,
                'events': ['*'],
                'url': 'https://example.com/',
                'repositories': [self.global_site_repo.pk],
            },
            limit_to_local_site=self.local_site)

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'repositories': ['Select a valid choice. 2 is not one of the '
                                 'available choices.'],
            })

    def test_with_localsite_in_data(self) -> None:
        """Testing WebHookTargetForm with a LocalSite in form data"""
        # Make sure the initial state and querysets are what we expect on init.
        form = WebHookTargetForm()

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repositories'].queryset),
                         [self.global_site_repo, self.local_site_repo])

        # Now test what happens when it's been fed data and validated.
        form = WebHookTargetForm(data={
            'apply_to': WebHookTarget.APPLY_TO_SELECTED_REPOS,
            'encoding': WebHookTarget.ENCODING_JSON,
            'events': ['*'],
            'url': 'https://example.com/',
            'local_site': self.local_site.pk,
            'repositories': [self.local_site_repo.pk],
        })

        self.assertIsNone(form.limited_to_local_site)
        self.assertIn('local_site', form.fields)
        self.assertEqual(list(form.fields['repositories'].queryset),
                         [self.global_site_repo, self.local_site_repo])

        self.assertTrue(form.is_valid())

        # Make sure any overridden querysets have been restored, so users can
        # still change entries.
        self.assertEqual(list(form.fields['repositories'].queryset),
                         [self.global_site_repo, self.local_site_repo])

        webhook = form.save()
        self.assertEqual(webhook.local_site, self.local_site)
        self.assertEqual(list(webhook.repositories.all()),
                         [self.local_site_repo])

    def test_with_localsite_in_data_and_instance(self) -> None:
        """Testing WebHookTargetForm with a LocalSite in form data and editing
        instance
        """
        webhook = WebHookTarget.objects.create()

        form = WebHookTargetForm(
            data={
                'apply_to': WebHookTarget.APPLY_TO_ALL,
                'encoding': WebHookTarget.ENCODING_JSON,
                'events': ['*'],
                'url': 'https://example.com/',
                'local_site': self.local_site.pk,
            },
            instance=webhook)
        self.assertTrue(form.is_valid())

        new_webhook = form.save()
        self.assertEqual(webhook.pk, new_webhook.pk)
        self.assertEqual(new_webhook.local_site, self.local_site)

    def test_with_localsite_in_data_and_invalid_repository(self) -> None:
        """Testing WebHookTargetForm with a LocalSite in form data and
        Repository not on the LocalSite
        """
        form = WebHookTargetForm(data={
            'apply_to': WebHookTarget.APPLY_TO_ALL,
            'encoding': WebHookTarget.ENCODING_JSON,
            'events': ['*'],
            'url': 'https://example.com/',
            'local_site': self.local_site.pk,
            'repositories': [self.global_site_repo.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'repositories': ['Select a valid choice. 2 is not one of the '
                                 'available choices.'],
            })

    def test_with_invalid_events(self) -> None:
        """Testing WebHookTargetForm with invalid events"""
        form = WebHookTargetForm(data={
            'apply_to': WebHookTarget.APPLY_TO_SELECTED_REPOS,
            'encoding': WebHookTarget.ENCODING_JSON,
            'events': ['review_request_closed', 'a', 'b'],
            'url': 'https://example.com/',
            'repositories': [self.global_site_repo.pk],
        })
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors,
            {
                'events': ['Select a valid choice. a is not one of the '
                           'available choices.'],
            })



"""Representations of remote repositories.

This is pending deprecation. Consumers should update their imports to use
the classes in :py:mod:`reviewboard.hostingsvcs.base.repository`.
"""

from reviewboard.hostingsvcs.base.repository import RemoteRepository


__all__ = [
    'RemoteRepository',
]


__autodoc_excludes__ = __all__



"""Tags for administrative functions."""

from __future__ import annotations

import re

from django import template
from django.contrib import messages
from django.contrib.admin.templatetags.admin_urls import (
    add_preserved_filters,
    admin_urlquote)
from django.urls import reverse
from django.utils.safestring import mark_safe
from djblets.util.templatetags.djblets_js import json_dumps_items

from reviewboard.admin.forms.change_form import ChangeFormFieldset


register = template.Library()


@register.simple_tag
def alert_css_classes_for_message(message):
    """Render the CSS classes for a rb-c-alert from a Django Message.

    This helps to craft an alert that reflects the status of a
    :py:class:`~django.contrib.messages.storage.base.Message`.

    This will include a CSS modifier class reflecting the status of the
    message and any extra tags defined on the message.

    Args:
        message (django.contrib.messages.storage.base.Message):
            The message to render classes for.

    Returns:
        unicode:
        A space-separated list of classes.
    """
    status_class = {
        messages.DEBUG: '-is-info',
        messages.INFO: '-is-info',
        messages.SUCCESS: '-is-success',
        messages.WARNING: '-is-warning',
        messages.ERROR: '-is-error',
    }[message.level]

    if message.extra_tags:
        return '%s %s' % (status_class, message.extra_tags)

    return status_class


@register.filter
def split_error_title_text(error):
    """Split an exception's text into a title and body text.

    Args:
        error (Exception):
            The error containing text to split.

    Returns:
        tuple:
        A tuple containing:

        1. The title text.
        2. The rest of the error message (or ``None``).
    """
    return str(error).split('\n', 1)


@register.simple_tag()
def process_result_headers(result_headers):
    """Process a Django ChangeList's result headers to aid in rendering.

    This will provide better information for our template so that we can
    more effectively render a datagrid.

    Args:
        result_headers (list of dict):
            The result headers to modify.
    """
    class_attrib_re = re.compile(r'\s*class="([^"]+)"')

    for header in result_headers:
        m = class_attrib_re.match(header['class_attrib'])

        if m:
            class_value = m.groups(1)[0]
        else:
            class_value = ''

        if class_value != 'action-checkbox-column':
            class_value = 'has-label %s' % class_value

        header['class_attrib'] = \
            mark_safe(' class="datagrid-header %s"' % class_value)

        if header['sortable'] and header['sort_priority'] > 0:
            if header['ascending']:
                sort_order = 'asc'
            else:
                sort_order = 'desc'

            if header['sort_priority'] == 1:
                sort_priority = 'primary'
            else:
                sort_priority = 'secondary'

            header['sort_icon'] = 'datagrid-icon-sort-%s-%s' % (
                sort_order, sort_priority)

    return ''


@register.simple_tag(takes_context=True)
def changelist_js_model_attrs(context):
    """Return serialized JSON attributes for the RB.Admin.ChangeListPage model.

    These will all be passed to the :js:class:`RB.Admin.ChangeListPage`
    constructor.

    Args:
        context (django.template.Context):
            The context for the page.

    Returns:
        django.utils.safestring.SafeText:
        A string containing the JSON attributes for the page model.
    """
    action_form = context.get('action_form')
    cl = context['cl']

    model_data = {
        'modelName': cl.opts.verbose_name,
        'modelNamePlural': cl.opts.verbose_name_plural,
    }

    if action_form is not None:
        action_choices = action_form.fields['action'].choices

        model_data['actions'] = [
            {
                'id': action_id,
                'label': action_label,
            }
            for action_id, action_label in action_choices
            if action_id
        ]

    return json_dumps_items(model_data)


@register.inclusion_tag('admin/submit_line.html', takes_context=True)
def change_form_submit_buttons(context):
    """Return HTML for a change form's submit buttons.

    This will compute the correct set of Save/Delete buttons, based on whether
    this is rendering for a Django admin change form (taking into account
    the object's state and user's permissions) or for any other type of form.

    Args:
        context (django.template.Context):
            The context for the page.

    Returns:
        django.utils.safestring.SafeText:
        A string containing the submit buttons.
    """
    show_save = context.get('show_save', True)
    delete_url = None

    if 'change' in context:
        change = context['change']
        is_popup = context['is_popup']
        show_delete = context.get('show_delete', True)

        if is_popup:
            show_delete = False
            show_save_as_new = False
            show_save_and_add_another = False
            show_save_and_continue = False
        else:
            save_as = context['save_as']
            opts = context['opts']
            original = context['original']

            show_delete = (
                change and
                context.get('show_delete', True) and
                context['has_delete_permission'])
            show_save_as_new = (
                save_as and
                change)
            show_save_and_add_another = (
                (not save_as or context['add']) and
                context['has_add_permission'])
            show_save_and_continue = (
                context.get('show_save_and_continue', True) and
                context['has_change_permission'])

            if show_delete:
                assert original is not None

                request = context['request']
                delete_urlname = '%s:%s_%s_delete' % (
                    request.current_app,
                    opts.app_label,
                    opts.model_name)

                delete_url = add_preserved_filters(
                    context,
                    reverse(delete_urlname,
                            args=[admin_urlquote(original.pk)]))
    else:
        delete_url = context.get('delete_url', '#')
        show_delete = context.get('show_delete', False)
        show_save_as_new = context.get('show_save_as_new', False)
        show_save_and_add_another = context.get('show_save_and_add_another',
                                                False)
        show_save_and_continue = context.get('show_save_and_continue', False)

    return {
        'delete_url': delete_url,
        'show_delete_link': show_delete,
        'show_save': show_save,
        'show_save_and_add_another': show_save_and_add_another,
        'show_save_and_continue': show_save_and_continue,
        'show_save_as_new': show_save_as_new,
    }


@register.filter
def change_form_fieldsets(admin_form):
    """Iterate through all fieldsets in an administration change form.

    This will provide each field as a
    :py:class:`~reviewboard.admin.forms.change_form.ChangeFormFieldset`.

    Args:
        admin_form (django.contrib.admin.helpers.AdminForm):
            The administration form.

    Yields:
        reviewboard.admin.forms.change_form.ChangeFormFieldset:
        Each fieldset in the form.
    """
    form = admin_form.form
    readonly_fields = admin_form.readonly_fields
    model_admin = admin_form.model_admin

    for name, options in admin_form.fieldsets:
        yield ChangeFormFieldset(form=form,
                                 name=name,
                                 readonly_fields=readonly_fields,
                                 model_admin=model_admin,
                                 **options)


@register.simple_tag(takes_context=True)
def render_change_form_fieldset(context, fieldset):
    """Render a Change Form fieldset.

    This will render a
    :py:class:`~reviewboard.admin.forms.change_form.ChangeFormFieldset` to
    HTML.

    Args:
        context (django.template.Context):
            The current template context.

        fieldset (reviewboard.admin.forms.change_form.ChangeFormFieldset):
            The fieldset to render.

    Returns:
        django.utils.safestring.SafeText:
        The resulting HTML for the fieldset.
    """
    return fieldset.render(context)



"""A hook for registering an authentication backend."""

from __future__ import annotations

from djblets.extensions.hooks import BaseRegistryHook, ExtensionHookPoint

from reviewboard.accounts.backends import auth_backends


class AuthBackendHook(BaseRegistryHook, metaclass=ExtensionHookPoint):
    """A hook for registering an authentication backend.

    Authentication backends control user authentication, registration, user
    lookup, and user data manipulation.

    This hook takes the class of an authentication backend that should
    be made available to the server.
    """

    registry = auth_backends

    def initialize(self, backend_cls):
        """Initialize the hook.

        This will register the provided authentication backend.

        Args:
            backend_cls (type):
                The authentication backend to register. This should be a
                subclass of
                :py:class:`~reviewboard.accounts.backends.AuthBackend`.
        """
        super(AuthBackendHook, self).initialize(backend_cls)



"""Views for the Review Board datagrids."""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING, Type

from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.http import Http404
from django.shortcuts import get_object_or_404, render
from django.utils.translation import gettext_lazy as _

from reviewboard.accounts.decorators import (check_login_required,
                                             valid_prefs_required)
from reviewboard.datagrids.grids import (DashboardDataGrid,
                                         GroupDataGrid,
                                         ReviewRequestDataGrid,
                                         UsersDataGrid,
                                         UserPageReviewsDataGrid,
                                         UserPageReviewRequestDataGrid)
from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.site.decorators import check_local_site_access
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse
    from djblets.datagrid.grids import DataGrid

    from reviewboard.site.models import LocalSite


#: The template used for rendering all datagrids.
#:
#: Version Added:
#:     5.0.7
_DATAGRID_TEMPLATE_NAME = 'datagrids/datagrid.html'


def _is_datagrid_gridonly(
    request: HttpRequest,
) -> bool:
    """Return whether or not the current request is for an embedded datagrid.

    This method allows us to disable consent checks in
    :py:func:`~reviewboard.accounts.decorators.valid_prefs_required` when a
    datagrid is requesting updated data so that we do not return a redirect
    and embed the result of that redirect in the datagrid instead.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

    Returns:
        bool:
        Whether or not this request is for an embedded datagrid.
    """
    return 'gridonly' in request.GET


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def all_review_requests(
    request: HttpRequest,
    *,
    local_site: Optional[LocalSite] = None,
) -> HttpResponse:
    """Display a list of all review requests.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site.

    Returns:
        django.http.HttpResponse:
        The rendered HTTP response for the datagrid.
    """
    datagrid = ReviewRequestDataGrid(
        request=request,
        queryset=ReviewRequest.objects.public(
            user=request.user,
            distinct=False,
            status=None,
            local_site=local_site,
            show_inactive=True),
        title=_('All Review Requests'),
        local_site=local_site)

    return datagrid.render_to_response(_DATAGRID_TEMPLATE_NAME)


@login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def dashboard(
    request: HttpRequest,
    local_site: Optional[LocalSite] = None,
) -> HttpResponse:
    """Display the dashboard.

    This shows review requests organized by a variety of lists, depending on
    the ``view`` GET parameter. Valid ``view`` parameters are:

    * ``incoming``
    * ``mine``
    * ``outgoing``
    * ``overview``
    * ``starred``
    * ``to-group``
    * ``to-me``

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site.

    Returns:
        django.http.HttpResponse:
        The rendered HTTP response for the datagrid. What datagrid is rendered
        depends on the ``view`` parameter.
    """
    grid = DashboardDataGrid(request=request,
                             local_site=local_site)

    return grid.render_to_response(_DATAGRID_TEMPLATE_NAME)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def group(
    request: HttpRequest,
    *,
    name: str,
    local_site: Optional[LocalSite] = None,
) -> HttpResponse:
    """Display a list of review requests belonging to a particular group.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        name (str):
            The name of the review group to view.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site.

    Returns:
        django.http.HttpResponse:
        The rendered HTTP response for the datagrid.
    """
    # Make sure the group exists
    group = get_object_or_404(Group, name=name, local_site=local_site)

    if not group.is_accessible_by(request.user):
        return render(request, 'datagrids/group_permission_denied.html',
                      status=403)

    datagrid = ReviewRequestDataGrid(
        request=request,
        queryset=ReviewRequest.objects.to_group(
            group_name=name,
            local_site=local_site,
            user=request.user,
            distinct=False,
            status=None),
        title=_('Review requests for %s') % group.display_name,
        local_site=local_site)

    return datagrid.render_to_response(_DATAGRID_TEMPLATE_NAME)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def group_list(
    request: HttpRequest,
    *,
    local_site: Optional[LocalSite] = None,
) -> HttpResponse:
    """Display a list of all review groups.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site.

    Returns:
        django.http.HttpResponse:
        The rendered HTTP response for the datagrid.
    """
    grid = GroupDataGrid(request=request,
                         local_site=local_site)

    return grid.render_to_response(_DATAGRID_TEMPLATE_NAME)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def group_members(
    request: HttpRequest,
    *,
    name: str,
    local_site: Optional[LocalSite] = None,
) -> HttpResponse:
    """Display a list of users registered for a particular group.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        name (str):
            The name of the review group to view.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site.

    Returns:
        django.http.HttpResponse:
        The rendered HTTP response for the datagrid.
    """
    # Make sure the group exists
    group = get_object_or_404(Group,
                              name=name,
                              local_site=local_site)

    if not group.is_accessible_by(request.user):
        return render(request, 'datagrids/group_permission_denied.html',
                      status=403)

    datagrid = UsersDataGrid(
        request=request,
        queryset=group.users.filter(is_active=True),
        title=_('Members of group %s') % name)

    return datagrid.render_to_response(_DATAGRID_TEMPLATE_NAME)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def submitter(
    request: HttpRequest,
    *,
    username: str,
    local_site: Optional[LocalSite] = None,
    grid: Optional[str] = None,
) -> HttpResponse:
    """Display a user's profile, showing their review requests and reviews.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site.

        grid (str):
            The name of the datagrid to view.

            This can be one of:

            * ``review-request``
            * ``reviews``

    Returns:
        django.http.HttpResponse:
        The rendered HTTP response for the datagrid.
    """
    # Make sure the user exists.
    if local_site:
        try:
            user = local_site.users.get(username=username)
        except User.DoesNotExist:
            raise Http404
    else:
        user = get_object_or_404(User, username=username)

    datagrid_cls: Type[DataGrid]

    if grid is None or grid == 'review-requests':
        datagrid_cls = UserPageReviewRequestDataGrid
    elif grid == 'reviews':
        datagrid_cls = UserPageReviewsDataGrid
    else:
        raise Http404

    datagrid = datagrid_cls(request, user, local_site=local_site)
    datagrid.tabs = [
        (UserPageReviewRequestDataGrid.tab_title,
         local_site_reverse('user',
                            local_site=local_site,
                            args=[username])),
        (UserPageReviewsDataGrid.tab_title,
         local_site_reverse('user-grid',
                            local_site=local_site,
                            args=[username, 'reviews']))
    ]

    return datagrid.render_to_response(_DATAGRID_TEMPLATE_NAME)


@check_login_required
@check_local_site_access
@valid_prefs_required(disable_consent_checks=_is_datagrid_gridonly)
def users_list(
    request: HttpRequest,
    *,
    local_site: Optional[LocalSite] = None,
) -> HttpResponse:
    """Display a list of all users.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site.

    Returns:
        django.http.HttpResponse:
        The rendered HTTP response for the datagrid.
    """
    grid = UsersDataGrid(request=request,
                         local_site=local_site)

    return grid.render_to_response(_DATAGRID_TEMPLATE_NAME)
